# coding: utf-8
"""
    Rgg.Spike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time, sleep
import logging
import asyncio
import numpy as np
from scipy.stats import multivariate_normal
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
from SliceUtils import slice_consecutives
from pomegranate import *
from molass_legacy.Peaks.ElutionModels import egh
from Prob.GaussianMixture import gaussian_pdf
import molass_legacy.KekLib.DebugPlot as plt
from .RggMixtureModel import RggMixtureModel
from .RggUtils import plot_histogram_2d
from molass_legacy.Peaks.ElutionModels import egh_pdf

def distributions_demo():
    x = np.arange(500)

    plt.push()
    fig, ax = plt.subplots()

    d = GaussianKernelDensity([200], bandwidth=10)
    ax.plot(x, d.probability(x))

    fig.tight_layout()
    plt.show()
    plt.pop()

    y = np.arange(100)
    xx, yy = np.meshgrid(x, y)
    plt.push()
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    cov = np.array([[900, 0], [0, 100]])
    rv = multivariate_normal([200, 50], cov)
    ax.plot_surface(xx, yy, rv.pdf(np.vstack([xx.flatten(), yy.flatten()]).T).reshape(xx.shape))

    fig.tight_layout()
    plt.show()
    plt.pop()

def generate_demo_data():
    x = np.arange(500)
    y = np.zeros(len(x))
    rg_list = []
    cy_list = []
    for rg, params in [
            # [0.6, 200, 40, 10],
            [40, [0.6, 200, 30,  5]],
            [30, [1.0, 300, 30, 10]],
            # [25, [0.4, 370, 30, 20]],
            ]:
        if True:
            cy = egh(x, *params)
        else:
            h, mu, sigma, tau = params
            cy = h*gaussian_pdf(x, mu, sigma)
        cy_list.append(cy)
        y += cy
        rg_list.append(rg)

    rg = np.sum(np.array(rg_list)[:,np.newaxis]*np.array(cy_list)/y, axis=0)
    return x, y, rg, len(cy_list)

def rgg_demo():
    from .RggEghDistribution import RggEghDistribution
    from RgProcess.RgCurve import convert_to_probabilitic_data
    x, y, rg, num_components = generate_demo_data()

    X = convert_to_probabilitic_data(x, y, rg)
    model = GeneralMixtureModel.from_samples(RggEghDistribution, num_components, X)
    print(model)

def spike_demo():
    x, y, rg, num_components = generate_demo_data()
    spike_demo_impl([(x, y, rg, None)], num_components=num_components)

def make_availability_slices(y, min_ratio):
    max_y = np.max(y)
    pairs = slice_consecutives(np.where(y/max_y >= min_ratio)[0])
    print("pairs=", pairs)
    slices = []
    states = []
    start = 0
    for f, t in pairs:
        if start < f:
            slices.append(slice(start, f))
            states.append(0)

        start = t+1
        slices.append(slice(f, start))
        states.append(1)
    if start < len(y):
        slices.append(slice(start, len(y)))
        states.append(0)
    return slices, states

PLOT_AS_SURFACE = True

def spike_demo_impl(data_list, num_components=None, use_peaks=False, seed=None, axes=None, plot_all=True, threed=True):
    from DataUtils import get_in_folder

    assert len(data_list) == 1

    if use_peaks:
        from molass_legacy.Peaks.RobustPeaks import RobustPeaks

    num_datasets = len(data_list)

    if axes is None:
        plt.push()
        fig = plt.figure(figsize=(21,6))
        gs = GridSpec(num_datasets, 3)
    else:
        fig = axes[0].figure

    fig.suptitle("2D EGH Mixture Model Demo for %s" % get_in_folder(), fontsize=20)

    fit_result = None
    for k, (x, y, rgc, X) in enumerate(data_list):
        max_y = np.max(y)

        model = RggMixtureModel(X=X, num_components=num_components)
        fit_result = model.fit()
        cy_list, ty = model.get_components(fit_result, x, y)

        if axes is None:
            ax1 = fig.add_subplot(gs[k,0])
            ax2 = fig.add_subplot(gs[k,1], projection='3d')
            projection = '3d' if threed else None
            ax3 = fig.add_subplot(gs[k,2], projection=projection)
        else:
            if plot_all:
                ax1, ax2, ax3 = axes
            else:
                ax3 = axes

        if plot_all:

            axt = ax1.twinx()
            axt.grid(False)

            ax1.set_title("Xray Elution and Rg Variation", fontsize=16)
            ax2.set_title("2D Histogram of Probability Variables", fontsize=16)
            random_text = "" if seed is None else " with seed=%d" % seed

            ax3.set_title("Decomposition Result%s" % random_text, fontsize=16)

            ax1.set_ylabel('Xray Intensity')
            ax1.set_xlabel('Eno')
            axt.set_ylabel('Rg')

            ax1.plot(x, y, color='orange')

            ax2.set_xlabel('Eno')
            ax2.set_ylabel('Rg')
            ax2.set_zlabel('Counts')
            ymin, ymax = ax1.get_ylim()
            segments = rgc.get_curve_segments()
            labeled = False
            for x_, y_, rg_ in segments:
                rg_label = None if labeled else 'Rg (smoothed)'
                axt.plot(x_, rg_, ':', color='C1', label=rg_label)
                rec_label = None if labeled else 'Rg-available range'
                p = Rectangle(
                        (x_[0], ymin),  # (x,y)
                        x_[-1] - x_[0], # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                        label=rec_label,
                    )
                ax1.add_patch(p)
                labeled = True

            ax1.legend()
            axt.legend()
            ymin, ymax = axt.get_ylim()
            axt.set_ylim(0, ymax*1.2)

            k = 0
            for slice_, state in zip(rgc.slices, rgc.states):
                if state == 0:
                    continue

                x_ = x[slice_]
                rg = segments[k][2]     # Rg
                if PLOT_AS_SURFACE:
                    plot_histogram_2d(ax2, x_, y[slice_], rg, max_y)
                else:
                    z_ = y[slice_]/max_y*100
                    b_ = np.zeros(len(z_))
                    ax2.bar3d(x_, rg, b_, 0.5, 0.05, z_/max_y*100, shade=True, edgecolor='green')
                k += 1

            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(10, ymax)

        if threed:
            ax3.set_xlabel('Eno')
            ax3.set_ylabel('Rg')
            ax3.set_zlabel('Intensity')
            for w, m, s, t, rg in fit_result.get_params_for_refiner():
                y_ =  w * egh_pdf(x, m, s, t)
                rg_ = np.ones(len(x))*rg
                plot_histogram_2d(ax3, x, y_, rg_, max_y)
            ax3.set_ylim(10, ymax)
        else:
            ax3.set_ylabel('Xray Intensity')
            ax3.set_xlabel('Eno')
            ax3.plot(x, y, color='orange')

            for k, cy in enumerate(cy_list):
                ax3.plot(x, cy, ':', label='component-%d' % (k+1))
            ax3.plot(x, ty, ':', color='red', label='total')
            ax3.legend()

    if axes is None:
        fig.tight_layout()
        plt.show()
        plt.pop()

    return fit_result

def get_data_info(in_folder, correction, both=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from RgProcess.RgCurve import RgCurve

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    if correction:
        sd = sp.get_corrected_sd()
    else:
        sd = sp.get_sd()
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    rgc = RgCurve(qv, xr_curve, D, E)
    # rgc.proof_plot()
    X = rgc.get_probabilistic_data()
    if both:
        return [D, E, qv, xr_curve, rgc, X], [sd.conc_array, sd.lvector, sd.get_uv_curve()]
    else:
        return D, E, qv, xr_curve, rgc, X

def spike_demo_real(in_folder, data_info=None, num_components=3, correction=True, use_peaks=False, seed=None, axes=None, refiner=True, plot_all=True):
    from molass_legacy._MOLASS.SerialSettings import set_setting

    logger = logging.getLogger(__name__)

    if data_info is None:
        D, E, qv, xr_curve, rgc, X = get_data_info(in_folder, correction)
    else:
        D, E, qv, xr_curve, rgc, X = data_info

    if seed is None:
        seed = np.random.randint(1000, 9999)
    np.random.seed(seed)
    print("seed=", seed)

    set_setting('in_folder', in_folder)
    fit_result = spike_demo_impl([[xr_curve.x, xr_curve.y, rgc, X]], num_components=num_components, use_peaks=use_peaks, seed=seed,
                                    axes=axes, plot_all=plot_all)

    logger.info("Mixture(%d) params=%s", seed, str(fit_result.get_params_for_refiner()))

    if refiner:
        from .RgRefiner import RgRefiner, plot_rgrefiner_result
        init_params = fit_result.get_params_for_refiner()
        print("init_params=", init_params)

        rr = RgRefiner(init_params)
        x_, y_, rg_ = rgc.get_valid_curves()
        rr.fit(x_, y_, rg_, debug=False)
        logger.info("Refiner(%d, %d) params=%s, func_value=%g", seed, rr.seed, str(rr.params), rr.func_value)

        plt.push()
        fig, ax = plt.subplots()
        x = xr_curve.x
        y = xr_curve.y
        plot_rgrefiner_result(ax, x, y, rgc, rr, seed)
        fig.tight_layout()
        plt.show()
        plt.pop()
    else:
        return rgc, fit_result

def spike_demo_many_trials(in_folder, num_components=3, plot_all=True, num_trials=1, mm_seed=None,rr_seeds=None, pause_at_turn=False):
    global finished

    print("num_trials=", num_trials)

    data_info, uv_info = get_data_info(in_folder, True, both=True)
    if False:
        from molass_legacy.Elution.CurveUtils import simple_plot
        D, E, qv, xr_curve, rgc, X = data_info
        U, wlvec, uv_curve = uv_info
        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,6))
        simple_plot(ax1, uv_curve)
        simple_plot(ax2, xr_curve, color='orange')
        fig.tight_layout()
        plt.show()
        plt.pop()

    for n in range(num_trials):

        dp = plt.push()
        fig  = plt.figure(figsize=(21,11))

        axes = np.empty((2,3), dtype=object)
        gs = GridSpec(2,3)
        for i in range(2):
            for j in range(3):
                projection = '3d'if plot_all and (i, j) in [(0,1), (0,2)] else None
                axes[i,j] = fig.add_subplot(gs[i,j], projection=projection)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9, hspace=0.2)
        fig.canvas.draw()
        dp.update()

        if plot_all:
            mm_seed_ = np.random.randint(1000, 9999) if mm_seed is None else mm_seed
            rgc, fit_result = spike_demo_real(in_folder, data_info, num_components, seed=mm_seed_, axes=axes[0,:], refiner=False, plot_all=plot_all)
            j = 0
            axes_ = axes[1,:]
            dp.after(100, lambda: refiner_loop(dp, j, axes_, mm_seed_, rr_seeds, rgc, fit_result, plot_all, pause_at_turn))
        else:
            assert False

        plt.show()
        plt.pop()

def refiner_loop(dp, j, axes, mm_seed, rr_seeds, rgc, fit_result, plot_all, pause_at_turn):
    from .RgRefiner import RgRefiner, plot_rgrefiner_result
    global finished

    logger = logging.getLogger(__name__)
    init_params = fit_result.get_params_for_refiner()
    print("init_params=", init_params)
    rr = RgRefiner(init_params)
    x_, y_, rg_ = rgc.get_valid_curves()
    min_overlap = True
    seed = None if rr_seeds is None else rr_seeds[j]
    rr.fit(x_, y_, rg_, seed=seed, min_overlap=min_overlap, debug=False)
    logger.info("Refiner(%d, %d) params=%s, func_value=%g", mm_seed, rr.seed, str(rr.params), rr.func_value)

    ax = axes[j] if plot_all else axes
    ecurve = rgc.ecurve
    plot_rgrefiner_result(ax, ecurve.x, ecurve.y, rgc, rr, mm_seed, min_overlap=min_overlap)
    dp.fig.canvas.draw()
    if j < 2:
        dp.after(100, lambda: refiner_loop(dp, j+1, axes, mm_seed, rr_seeds, rgc, fit_result, plot_all, pause_at_turn))
    else:
        sleep(1)
        path = get_figure_path()
        dp.fig.savefig(path)
        if pause_at_turn:
            pass
        else:
            dp.cancel()

fig_no = -1
def get_figure_path():
    from molass_legacy._MOLASS.SerialSettings import get_setting
    global fig_no
    while True:
        fig_no += 1
        path  = os.path.join(get_setting("temp_folder"), "fig-%03d.jpg" % fig_no)
        if os.path.exists(path):
            continue
        else:
            break
    return path

def mm_refiner_combined_loop():
    pass

async def get_content(n):
    await asyncio.sleep(3)
    return n + 1

async def f(n):
    content = await get_content(n)
    return content

def coroutine_spike():
    loop = asyncio.get_event_loop()
    v = loop.run_until_complete(f(1))
    print(v)



# coding: utf-8
"""
    ProofPlot.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""
import re
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.ticker      import FuncFormatter
from bisect                 import bisect_right
from PlotUtils              import get_wider_range, convert_to_the_level
from molass_legacy.KekLib.OurMatplotlib          import convert_to_fixed_font_latex

DEBUG = False

class Plotter:
    def __init__( self, intensity, title, rg_info_a=None, rg_info_k=None, orig_info=None, q_limit_index=None ):
        print('Plotter: q_limit_index=', q_limit_index)

        self.orig_array = intensity.orig_array
        # self.x      = intensity.X
        # self.y      = intensity.Y
        self.x = self.orig_array[:,0]**2
        self.y = np.log( self.orig_array[:,1] )

        if rg_info_k is None:
            self.sx     = None
            self.sy     = None
        else:
            dx  = 1e-4
            self.sx = np.arange( self.x[0], self.x[-1] + 0.5*dx, dx )
            x_  = np.sqrt( self.sx )
            # y_  = rg_info_k.fit.model( x_ )
            # self.sy = np.log( y_ )
            self.sy = self.y

        self.rg_info_a  = rg_info_a
        self.rg_info_k  = rg_info_k

        self.fig    = fig = plt.figure( figsize=( 24, 8 ) )
        self.ax1    = fig.add_subplot( 1, 3, 1 )
        self.ax2    = fig.add_subplot( 1, 3, 2 )
        self.ax3    = fig.add_subplot( 1, 3, 3 )

        t_ = len( self.sx ) // 8
        self.sxmin  = self.sx[0]
        self.sxmax  = self.sx[t_]
        sy_ = self.sy[ 0:t_+1 ]
        sy__ = sy_[np.isfinite(sy_)]
        self.symin  = np.min( sy__ )
        self.symax  = np.max( sy__ )

        self.ax1.set_title( title )
        self.draw_scatter_and_lines( self.ax1, orig_info=orig_info )

        sxmin_, sxmax_ = get_wider_range( self.sxmin, self.sxmax, 0.2 )
        symin_, symax_ = get_wider_range( self.symin, self.symax, 0.2 )

        self.ax2.set_title( title + ' (zoomed-in)' )
        self.ax2.set_xlim( sxmin_, sxmax_ )
        self.ax2.set_ylim( symin_, symax_ )
        self.draw_scatter_and_lines( self.ax2, curvature=True )

        self.ax3.set_title( title + ' (zoomed-in closer)' )

        sxmin3_, sxmax3_ = sxmin_, sxmax_ 
        if DEBUG: print( '(1) sxmin3_, sxmax3_=', [sxmin3_, sxmax3_] )
        symin3_, symax3_ = symin_, symax_

        if rg_info_a is not None:
            f0, t0  = rg_info_a.From, rg_info_a.To
            y_ = self.y[ f0:t0+1 ]
            sxmin3_, sxmax3_ = get_wider_range( self.x[f0], self.x[t0], 0.1 )
            symin3_, symax3_ = get_wider_range( np.min( y_ ), np.max( y_ ), 0.1 )

        if rg_info_k is not None:
            f2, t2  = rg_info_k.From, rg_info_k.To
            if f2 is None or t2 is None: return
            y_ = self.y[ f2:t2+1 ]
            xmin_, xmax_ = get_wider_range( self.x[f2], self.x[t2], 0.1 )
            y__ = y_[np.isfinite(y_)]
            ymin_, ymax_ = get_wider_range( np.min( y__ ), np.max( y__ ), 0.1 )
            if rg_info_a is None:
                sxmin3_ = xmin_
                sxmax3_ = xmax_
                symin3_ = ymin_
                symax3_ = ymax_
            else:
                sxmin3_ = min( sxmin3_, xmin_ )
                sxmax3_ = max( sxmax3_, xmax_ )
                symin3_ = min( symin3_, ymin_ )
                symax3_ = max( symax3_, ymax_ )
        if DEBUG: print( '(2) sxmin3_, sxmax3_=', [sxmin3_, sxmax3_] )


        sxmin3_, sxmax3_ = get_wider_range( sxmin3_, sxmax3_, 0.25 )
        if DEBUG: print( '(3) sxmin3_, sxmax3_=', [sxmin3_, sxmax3_] )
        symin3_, symax3_ = get_wider_range( symin3_, symax3_, 0.25 )

        self.ax2.plot(  [ sxmin3_, sxmax3_, sxmax3_, sxmin3_, sxmin3_ ],
                        [ symin3_, symin3_, symax3_, symax3_, symin3_ ],
                        ':', color='black' )

        self.ax3.set_xlim( sxmin3_, sxmax3_ )
        self.ax3.set_ylim( symin3_, symax3_ )
        self.draw_scatter_and_lines( self.ax3, curvature=True, marker=True )

        self.get_qrg_min_max( [ rg_info_a, rg_info_k ] )

        if rg_info_a is not None:
            self.draw_guineier( rg_info_a, 'red' )
            # self.plot_q_rg( rg_info_a, 'red' )

        if rg_info_k is not None:
            self.draw_guineier( rg_info_k, 'cyan' )
            # self.plot_q_rg( rg_info_k, 'cyan' )

        if q_limit_index is not None:
            # fix ax1 limits
            if q_limit_index < self.orig_array.shape[0]:
                xmin1, xmax1 = self.ax1.get_xlim()
                self.ax1.set_xlim( xmin1, xmax1 )
                ymin1, ymax1 = self.ax1.get_ylim()
                self.ax1.set_ylim( ymin1, ymax1 )

                # draw
                q = self.orig_array[q_limit_index,0]
                x = q**2
                self.ax1.plot( [ x, x ], [ ymin1, ymax1 ], color='gray' )

    def x_formatter( self, val, pos ):
        if pos is None: return ''

        if val >= 0:
            index = bisect_right( self.x, val )
            return '%g\n(%g)' % ( val, index )
        else:
            return '%g' % val

    def draw_scatter_and_lines( self, ax, curvature=False, marker=False, orig_info=None ):
        if orig_info is None:
            x_  = self.x
            y_  = self.y
        else:
            sd, n =orig_info
            x   = sd.intensity_array[n,:,0]
            y   = sd.intensity_array[n,:,1]
            positive = y > 0
            x_  = x[positive]**2
            y_  = np.log( y[positive] )

        ax.scatter( x_, y_, s=16 )

        if orig_info is not None:
            ymin, ymax = ax.get_ylim()
            ax.set_ylim( ymin, ymax )
            q_index = sd.get_usable_q_limit()
            if q_index is not None:
                x_  = x[q_index]**2
                ax.plot( [ x_, x_ ], [ ymin, ymax ], color='gray' )

        marker_ = 'x' if marker else None
        # ax.plot( self.sx, self.sy, color='pink', marker=marker_, label='Guinier-Porod fitted curve' )

        ax.set_xlabel( 'Q**2' )
        ax.set_ylabel( 'Ln( I )' )
        ax.xaxis.set_major_formatter( FuncFormatter( self.x_formatter ) )

        sxmin   = self.sxmin
        sxmax   = self.sxmax
        symin   = self.symin
        symax   = self.symax
        ax.plot(    [ sxmin, sxmax, sxmax, sxmin, sxmin ],
                    [ symin, symin, symax, symax, symin ],
                    ':', color='black' )

        if curvature:
            self.draw_curvature_etc( ax, marker=marker )

    def draw_curvature_etc( self, ax, marker=False ):
        axmin_, axmax_ = ax.get_xlim()
        aymin_, aymax_ = ax.get_ylim()

        xoffset = ( axmax_ - axmin_ ) * 0.05
        yoffset = ( aymax_ - aymin_ ) * 0.05

        # Guinier-Porod model boundaries
        if self.rg_info_k is not None and self.rg_info_k.fit.Rg is not None:
            # Rg, d = self.rg_info_k.fit.Rg, self.rg_info_k.fit.degree
            Rg, d = 99, 3
            for d in range( 1, 5 ):
                q = 1/Rg * np.sqrt( 3*d/2 )
                x = q**2
                label_ = 'Guinier-Porod Model boundaries' if d == 1 else None
                ax.plot( [ x, x ], [ aymin_, aymax_ ], '-', color='yellow', label=label_ )

        if self.rg_info_k is not None:
            q1  = self.rg_info_k.fit.q1
            x1  = q1**2
            x_  =  np.array( [q1] )
            y_  = np.log( self.rg_info_k.fit.model( x_ ) )
            y1  = y_[0]
            text_ = 'q1=%.3g, q1*Rg=%.3g' % ( q1, q1 * self.rg_info_k.fit.Rg )
            ax.annotate( text_, xy=( x1, y1 ),
                xytext=( x1 + 2*xoffset, y1 - yoffset ), alpha=0.5,
                arrowprops=dict( headwidth=5, width=1, color='black', shrink=0 ),
                )

    def draw_guineier( self, rec, color ):
        vstrs = []
        for v in [ rec.Rg, rec.Rg_stdev ]:
            vstr = 'NA' if v is None else '%.2f' % v
            vstrs.append( vstr )

        label = '%s: Rg=%s, Rg.stdev=%s' % ( rec.type, vstrs[0], vstrs[1] )

        for i, ax in enumerate( [ self.ax1, self.ax2, self.ax3 ] ):
            label_ = convert_to_fixed_font_latex( label )
            self.plot_guineier_inverval( ax, rec, color, label=label_ )

    def plot_guineier_inverval( self, ax, rec, color, label=None ):
        x = self.x
        y = self.y

        Rg, I0, f, t, a_, b_ = rec.Rg, rec.I0, rec.From, rec.To, rec.a, rec.b

        # print( 'plot_guineier_inverval:', a_,  - ( Rg**2 / 3 ), b_, np.log( I0 ) )
        if Rg is None: return

        if not np.allclose( [a_, b_], [ -( Rg**2 / 3 ), np.log( I0 ) ] ):
            print( 'WARNINIG: not allclose!', [a_, b_], [ -( Rg**2 / 3 ), np.log( I0 ) ] )

        b_ = np.log( I0 )
        a_ = - ( Rg**2 / 3 )
        y0 = b_ + a_ * x[f]
        y1 = b_ + a_ * x[t]

        ax.plot( [ x[f], x[t] ], [ y0, y1 ], marker='o', color=color, label=label )

    def get_qrg_min_max( self, rec_array ):
        self.sqrt_x = np.sqrt( self.x )

        qrg_array = [ 0, 2.46 ]

        for rec in rec_array:
            if rec is None: continue
            if rec.Rg is None: continue
            Rg, f, t = rec.Rg, rec.From, rec.To
            for x in self.sqrt_x[f:t+1]:
                qrg_array.append( x * Rg )

        self.qRg_min = np.min( qrg_array )
        self.qRg_max = np.max( qrg_array )

    def plot_q_rg( self, rec, color ):
        # print( 'plot_q_rg: color=', color, ', rec=', rec )

        ax = self.ax3
        label = '%s: q*Rg' % rec.type
        x = self.sqrt_x

        Rg, f, t = rec.Rg, rec.From, rec.To
        if Rg is None: return

        x_ = x[f:t+1]
        qRg =  x_ * Rg
        ymin, ymax = ax.get_ylim()
        qRg_limits = [ 0.8, 1.3 ]
        qRg_ = np.array( list( qRg ) + qRg_limits + [ 0, 1.7 ] )
        qRg_min = self.qRg_min
        qRg_max = self.qRg_max
        qRg_plot_max = 0.5
        y_ = convert_to_the_level( qRg, ymin, ymax, 0.0, qRg_plot_max, vmin=qRg_min, vmax=qRg_max )
        x__ = x_ ** 2
        label_ = convert_to_fixed_font_latex( label )
        ax.plot( x__, y_, ':',  color=color, label=label_ )

        xmin, xmax = ax.get_xlim()
        xoffset = ( xmax - xmin ) * 0.05
        yoffset = ( ymax - ymin ) * 0.05

        qRg_limits_ = convert_to_the_level( qRg_limits, ymin, ymax, 0.0, qRg_plot_max, vmin=qRg_min, vmax=qRg_max )

        for k, limit in enumerate( qRg_limits_ ):
            # print( 'plot_q_rg: color=', color, ', limit=', limit )
            ax.plot( [ xmin, xmax ], [ limit, limit ], ':', color='black' )
            text_ = 'q*Rg=%g' % qRg_limits[k]
            ax.annotate( text_, xy=( xmin, limit ),
                            xytext=( xmin + xoffset, limit + yoffset ),
                            color='black', alpha=0.5,
                            arrowprops=dict( headwidth=5, width=1, color='black', shrink=0.05 ),
                            )

        y__ = convert_to_the_level( [ qRg[0], qRg[-1] ], ymin, ymax, 0.0, qRg_plot_max, vmin=qRg_min, vmax=qRg_max )
        for i, k in enumerate( [ 0, -1 ] ):
            x__ = x_[k]**2
            text_ = 'q*Rg=%.3g' % qRg[k]
            xoffset_ = xoffset if i==0 else -xoffset*2
            sign_ = 1 if color == 'red' else -1
            yoffset_ = sign_ * ( yoffset if i==0 else -yoffset )
            color_ = color if color == 'red' else 'black'
            ax.annotate( text_, xy=( x__, y__[i] ),
                            xytext=( x__ + xoffset_, y__[i] + yoffset_ ),
                            color=color_, alpha=0.5,
                            arrowprops=dict( headwidth=5, width=1, color=color, shrink=0.05 ),
                            )

    def show_prepare( self ):
        # this is to avoid plt.show() when used with tkinter.
        for ax in [ self.ax1, self.ax2, self.ax3 ]:
            ax.legend()
        # suppressed for the latest version of matplotlib
        # self.fig.tight_layout()

    def show( self, block=True ):
        self.show_prepare()
        plt.show( block )




"""
    Batch.BatchUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, set_setting
from molass_legacy.Batch.LiteBatch import LiteBatch     

def load_lrfsrc(in_folder, clear=True):
    if clear:
        clear_temporary_settings()
    set_setting('in_folder', in_folder)
    batch = LiteBatch()
    lrf_src = batch.get_lrf_source(in_folder=in_folder, clear=False)
    return lrf_src

def load_v2_result(job_folder):
    from molass_legacy.Optimizer.FullOptResult import FullOptResult

    in_data_info_txt = os.path.join(job_folder, "in_data_info.txt")
    with open(in_data_info_txt) as fh:
        for line in fh:
            in_folder = line.strip().split("=")[-1]
    print("in_folder=", in_folder)
    lrfsrc = load_lrfsrc(in_folder)
    sd = lrfsrc.sd
    print("type(sd)=", type(sd))
    return FullOptResult(sd, lrfsrc.pre_recog, job_folder, set_folder_setting=True)


"""
    SecTheory.MartinSynge.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import gaussian

def num_plates_demo():
    tR = 500
    tI = 0
    sigma = 10
    x = np.arange(1000)

    def compute_num_plates(tR, sigma):
        return (tR/sigma)**2

    def compute_pdf(tR, sigma):
        y = gaussian(x, 1, tR, sigma)
        y /= np.sum(y)
        return y

    N = compute_num_plates(tR, sigma)
    y = compute_pdf(tR, sigma)

    init_params = [tR, sigma]
    slider_params = init_params.copy()

    def set_title(ax, N):
        ax.set_title(r"Theoretical Number of Plates: $ N = (\frac{t_R}{\sigma})^2 = %d $" % round(N), fontsize=20)

    with plt.Dp():
        fig, ax = plt.subplots(figsize=(12, 5))
        set_title(ax, N)
        ax.set_xlabel("Time (Frames)")
        ax.set_ylabel("Density")
        curve, = ax.plot(x, y)
        ax.axvline(x=0, label='$t_I$', color='gray')
        line = ax.axvline(x=tR, label='$t_R$', color='red')
        ax.legend()
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, 0.1)
        fig.tight_layout()
        fig.subplots_adjust(right=0.7)

        slider_specs = [    (r"$t_R$", 0, 1000, slider_params[0]),
                            (r"$\sigma$", 1, 50, slider_params[1]),
                            ]

        def slider_update(k, val):
            # print([k], "slider_update", val)
            slider_params[k] = val
            tR, sigma = slider_params
            line.set_xdata(tR)
            y = compute_pdf(tR, sigma)
            curve.set_data(x, y)
            N = compute_num_plates(tR, sigma)
            set_title(ax, N)
            fig.canvas.draw_idle()

        left = 0.75
        width = 0.2
        slider_axes = []
        sliders = []

        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([left, 0.8 - 0.08*k, width, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = fig.add_axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        plt.show()


"""
    Models.Stochastic.PaperDondi2002.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_theme()

def demo_monopore():
    from SecTheory.BasicModels import robust_single_pore_pdf_scaled
    from SecTheory.SecCF import gec_monopore_phi, dispersive_monopore
    from SecTheory.SecPDF import FftInvPdf
    from HdcTheory.ElutionCurve import compute_elution_curve, NUM_PLATES_PC, C, retention_time
    monopore_pdf = FftInvPdf(gec_monopore_phi)
    dispersive_monopore_pdf = FftInvPdf(dispersive_monopore)

    x = np.arange(300, 800)
    t0 = 400
    N = 1000
    T = 0.5
    poresize = 75
    rgs = [150, 90, 50, 35, 20]
    rv = np.linspace(10, poresize, 100)
    rhov = rv/poresize
    rhov[rhov > 1] = 1
    m = 3
    trv = t0 + N*T*(1 - rhov)**m
    r0 = 5000
    rv_hdc = np.linspace(poresize, 200, 50)
    sigma0 = 10
    tp = np.sqrt(NUM_PLATES_PC)*sigma0
    tI = t0 - tp
    trv_hdc = tI + retention_time(tp, rv_hdc, r0, C)
    fig, axes = plt.subplots(nrows=len(rgs), ncols=2,figsize=(18,10))
    for axrow, rg in zip(axes, rgs):
        for k, ax in enumerate(axrow):
            ax.set_xlabel("Time (frames)")
            ax.set_ylabel("Density")
            axt = ax.twinx()
            axt.grid(False)
            axt.set_ylabel("$R_g$")
            axt.plot(trv, rv, color="yellow", label=r"Exclusion Curve: $t_0 + N T (1 - \rho)^m$")
            if rg > 100:
                axt.plot(trv_hdc, rv_hdc, color="cyan", label=r"Calibration Curve: $T_0/(1 + 2 \lambda - C \lambda^2)$")
            axt.legend(loc="upper center", fontsize=9)
            if rg <= 100:
                rg_str = "$R_g$=%g" % rg if rg < poresize else "$R_g$ > %g" % poresize
                ax.set_title("Monopore Dispersive Model Elution Curves with varying $N_0$ with N=%g, T=%g, Poresize=%g, %s" % (N, T, poresize, rg_str))
                
            if k == 0:
                ax.set_ylim(-0.01, 0.21)
            if rg < poresize:
                rho = rg/poresize
                ni_ = N*(1 - rho)**1.5
                ti_ = T*(1 - rho)**1.5

                y = robust_single_pore_pdf_scaled(x - t0, ni_, ti_)
                y_ = y/np.sum(y)

                # y2 = monopore_pdf(x, ni_, ti_, t0)
                # y3 = monopore_pdf(x - t0, ni_, ti_, 0)
            
                ax.plot(x, y_, label='Non-Dispersive (i.e., $N_0=\infty$)')
                # ax.plot(x, y2, ":", label='CF inversion')
                # ax.plot(x, y3, label='CF inversion')
            else:
                ni_ = 0
                ti_ = 0

            if rg > 100:
                sigma = 10
                ax.set_title(r"Hydrodynamic Model Elution Curves with $R_g$ > Poresize=%g $r_0$=%g, $\sigma_0$=%g" % (poresize, r0, sigma))
                for j, rg_ in enumerate([200, 150, 100]):
                    y5 = compute_elution_curve(x, t0, rg_, r0, sigma)
                    y5_ = y5/np.sum(y5)
                    ax.plot(x, y5_, color="C%d" % (4 + j), label='HDC $R_g$=%g' % rg_)
            else:
                for k, N0 in enumerate([1600, 2500, 3600]):
                    y4 = dispersive_monopore_pdf(x, ni_, ti_, N0, t0)
                    ax.plot(x, y4, color="C%d" % (k+1), label='Dispersive $N_0$=%g' % N0)
            ax.axvline(x=t0, color='red', label='t0')

            ax.legend(fontsize=9)

    fig.tight_layout()
    plt.show()


if __name__ == "__main__":
    import sys
    # import seaborn as sns
    # sns.set_theme()
    sys.path.append("../lib")

    demo_monopore()


"""
    Models/Stochastic/LognormalPoreColumn.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize, basinhopping
import time
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.Models.ElutionModelUtils import compute_4moments
from Experiment.ColumnTypes import get_all_poresizes
from molass_legacy.Models.Stochastic.SecModelUtils import (
    DEFAULT_SIGMA, SIGMA_LOWER_BOUND,
    NUM_ENTRIES_LIST, DEFAULT_NUM_ENTRIES, DEFAULT_ENTRY_TIME,
    compute_tRv
    )
from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_cf, lognormal_pore_func, distr_func

def compute_residual_ratio_impl(x, y, rgs, params, debug=False):
    cy_list = []
    for rg, scale in zip(rgs, params[7:]):
        cy = lognormal_pore_func(x, scale, params[0], params[1], params[3], params[4], params[5], params[6], rg, params[2])
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    ratio = np.sum(np.abs(y - ty))/np.sum(y)

    if debug:
        print("ratio=", ratio)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":")
            ax.plot(x, ty, ":", color="red")
            fig.tight_layout()
            plt.show()

    return ratio

def plot_state_impl(x, y, title, cy_list, ty, mu, sigma, save_fig_as=None):
    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle(title, fontsize=20)
        ax1.set_title("Elution Curve", fontsize=16)
        ax2.set_title("Pore Size Distribution", fontsize=16)
        ax1.plot(x, y)
        for cy in cy_list:
            ax1.plot(x, cy, ":")
        ax1.plot(x, ty, ":", color="red")
        mode = np.exp(mu - sigma**2)
        max_size = int(mode + 200)
        z = np.arange(max_size)
        ax2.plot(z, distr_func(z, mu, sigma), label="PSD: Lognormal(%.3g, %.3g)" % (mu, sigma))
        xmin, xmax = ax2.get_xlim()
        ymin, ymax = ax2.get_ylim()
        tx = xmin*0.3 + xmax*0.7
        ty = ymin*0.5 + ymax*0.5
        poresize = np.exp(mu - sigma**2)
        ax2.text(tx, ty, "Poresize = %.3g" % poresize, ha="center", alpha=0.5, fontsize=16)
        ax2.legend()            
        fig.tight_layout()
        if save_fig_as is not None:
            plt.show(block=False)
            fig.savefig(save_fig_as)
            ret = True
        else:
            ret = plt.show()
    return ret

class LognormalPoreColumn:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.N = DEFAULT_NUM_ENTRIES
        self.T = DEFAULT_ENTRY_TIME
        self.x0 = 0
        self.me = 1.5
        self.mp = 1.5
        self.mu = 4.5
        self.sigma = DEFAULT_SIGMA
        self.evaluate_area_props = False
        self.abort = False

    def get_column_params(self, T=None, x0=None, mu=None, sigma=None):
        if T is None:
            T = self.T
        if x0 is None:
            x0 = self.x0
        if mu is None:
            mu = self.mu
        if sigma is None:
            sigma = self.sigma
        return np.array([self.N, T, x0, self.me, self.mp, mu, sigma])

    def prepare_for_rgs(self, x, y, rg_recs, debug=False):    
        self.init_scales = None
        W, M1, M2, M3 = compute_4moments(x, y)
        self.M1 = M1
        s = np.sqrt(M2)
        self.x0 = M1 - 8*s
        self.data_width = 3*s
        print("x0=%.3g, data_width=%.3g" % (self.x0, self.data_width))
        self.area = W

        self.rgs = rg_recs[:,1]
        max_rg = np.max(self.rgs)

        poresizes = np.asarray(get_all_poresizes())
        possible_poresizes = poresizes[poresizes > max_rg + 20]
        self.logger.info("trying for rgs=%s against possible_poresizes: %s", str(self.rgs), str(possible_poresizes))

        ret_list = []
        for poresize in possible_poresizes:
            ret = self.prepare_for_poresize(x, y, rg_recs, poresize, debug=False)
            if ret is None:
                return False
            ret_list.append(ret)

        # select the best poresize
        ret_array = np.asarray(ret_list)
        fv_list = ret_array[:,0]
        j = np.argmin(fv_list)
        self.poresize = possible_poresizes[j]
        self.x0 = ret_array[j,1]
        self.N = ret_array[j,2]
        self.logger.info("selected poresize=%.3g with x0=%.3g from evaluations %s", self.poresize, self.x0, str([round(v) for v in fv_list]))
        if debug:
            fv = fv_list[j]
            m = self.me + self.mp
            t_rgs = rg_recs[:,0]
            tRv = compute_tRv(self.N, self.T, self.x0, m, self.poresize, self.rgs)
            ret = self.plot_exclusion_state(x, y, self.poresize, self.rgs, self.x0, t_rgs, tRv, fv, title="selected exclusion state")
            if not ret:
                return False
        self.estimate_mu_sigma_from_poresize(self.poresize, debug=debug)
        return True

    def prepare_for_poresize(self, x, y, rg_recs, poresize, fit_optimize=False, debug=False):
        self.logger.info("preparing for poresize(%.3g)", poresize)
        rgs = rg_recs[:,1]

        m = self.me + self.mp
        t_rgs = rg_recs[:,0]
        def prepare_objective(p, debug=False, title=None):
            tRv = compute_tRv(self.N, self.T, p[0], m, poresize, rgs)
            # fv = np.sum((np.mean(tRv) - self.M1)**2) + ((tRv[-1] - tRv[0]) - self.data_width)**2
            fv = np.sum((tRv - t_rgs)**2)
            if debug:
                ret = self.plot_exclusion_state(x, y, poresize, rgs, p[0], t_rgs, tRv, fv, title=title)
                if not ret:
                    self.abort = True
            return fv

        fv_list = []
        res_list = []
        for N in NUM_ENTRIES_LIST:
            self.N = N
            if debug:
                fv = prepare_objective([self.x0], debug=True, title="before minimize with N=%d" % N)
                if self.abort:
                    return
            # bounds = [(-100, 100)]
            bounds = None
            res = minimize(prepare_objective, [self.x0], bounds=bounds, method="Nelder-Mead")
            if debug:
                fv = prepare_objective([res.x[0]], debug=True, title="after minimize with N=%d" % N)
                if self.abort:
                    return

            if res.success:
                self.logger.info("x0 = %f" % res.x[0])
            else:
                self.logger.error("x0 fitting failed")
                return

            if fit_optimize:
                x0 = res.x[0]
                res = self.prepare_fit_optimize(x, y, N, x0, rgs, debug=debug)
                if res is None:
                    return

            fv = res.fun
            self.logger.info("fv = %.3g" % fv)
            fv_list.append(fv)
            res_list.append(res)

        j = np.argmin(fv_list)
        res = res_list[j]
        min_fv = res.fun
        N = NUM_ENTRIES_LIST[j]
        self.logger.info("min_fv=%.3g, x0=%.3g for N=%d", min_fv, res.x[0], N)
        return min_fv, res.x[0], N

    def plot_exclusion_state(self, x, y, poresize, rgs, x0, t_rgs, tRs, fv, title):
        print("rgs=", rgs)
        print("t_rgs=", t_rgs)
        print("tRs=", tRs)
        m = self.me + self.mp
        rv = np.linspace(20, poresize, 100)
        plot_params = [self.N, self.T, x0, m, poresize].copy()
        slider_specs = [    ("N", 0, 6000, plot_params[0]),
                            ("T", 0, 3, plot_params[1]),
                            ("x0",  -500, 500, plot_params[2]),
                            ("m", 0, 4, plot_params[3]),
                            ("poresize", 0, 500, plot_params[4]),
                            ]
        with plt.Dp():
            fig, ax = plt.subplots(figsize=(15,5))
            if title is not None:
                ax.set_title(title + "; fv=%.3g" % fv, fontsize=20)
            ax.plot(x, y)
            ax.axvline(x0, color="red")
            axt = ax.twinx()
            axt.grid(False)
            excl_curve, = axt.plot(compute_tRv(*plot_params, rv), rv, color="yellow")
            axt.plot(t_rgs, rgs, "o", color="red")
            axt.plot(tRs, rgs, "o", color="green")

            dp = plt.get_dp()
            def slider_update(k, val):
                # print([k], "slider_update", val)
                plot_params[k] = val
                trv = compute_tRv(*plot_params, rv)
                excl_curve.set_data(trv, rv)
                dp.draw()

            slider_axes = []
            sliders = []
            for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                ax_ = fig.add_axes([0.75, 0.8 - 0.08*k, 0.2, 0.03])
                slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                slider_axes.append(ax_)
                sliders.append(slider)

            fig.tight_layout()
            fig.subplots_adjust(right=0.65)
            ret = plt.show()
        return ret

    def prepare_fit_optimize(self, x, y, N, x0, rgs, debug=False):

        def prepare_fit_objective(p, debug=False, title=None):
            cy_list = []
            for scale, rg in zip(p, rgs):
                cy = self.compute_curve(x, x0, scale, rg)
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            fv = np.sum((ty - y)**2)
            if debug:
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.plot(x, y)
                    for cy in cy_list:
                        ax.plot(x, cy, ":")
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout()
                    ret = plt.show()
                if not ret:
                    self.abort = True
            return fv

        num_components = len(rgs)
        init_scales = np.ones(num_components)*self.area/num_components
        if debug:
            fv = prepare_fit_objective(init_scales, debug=True, title="before fit minimize with N=%d" % N)
            if self.abort:
                return
        bounds = [(0, self.area*2)]*num_components
        res = minimize(prepare_fit_objective, init_scales, method="Nelder-Mead", bounds=bounds)
        if debug:
            fv = prepare_fit_objective(res.x, debug=True, title="after fit minimize with N=%d" % N)
            if self.abort:
                return

        return res

    def estimate_mu_sigma_from_poresize(self, poresize, debug=False):
        self.sigma = DEFAULT_SIGMA
        self.mu = np.log(poresize) + self.sigma**2      # note that: poresize = mode = exp(mu - sigma**2) 
        self.logger.info("estimated mu and sigma as %.3g, %.3g from poresize(%.3g)", self.mu, self.sigma, poresize)

    def estimate_optimal_ksec(self, x, y, rg_recs, debug=False):
        rgs = rg_recs[:,1]
        self.x = x
        self.y = y

        def ksec_objective(p, debug=False, title="ksec_objective"):
            T = p[0]
            x0 = p[1]
            cy_list = []
            for rg, scale in zip(rgs, p[2:]):
                cy = lognormal_pore_func(x, scale, self.N, T, self.me, self.mp, self.mu, self.sigma, rg, x0)
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            fv = np.sum((ty - y)**2)
            if debug:
                print("T=", T)
                lnp_params = np.concatenate([self.get_column_params(T=T, x0=x0), p[2:]])
                self.plot_state(title, lnp_params)
            return fv

        num_components = len(rgs)
        init_scale = self.area/num_components
        init_params = [self.T, self.x0] + [init_scale]*num_components
        if debug:
            ksec_objective(init_params, debug=True, title="before minimize ksec_objective")
            if self.abort:
                return False
        t0 = time.time()
        bounds = [(0.01, 10), (-1000, 1000)] + [(0, 10)]*num_components
        ret = minimize(ksec_objective, init_params, bounds=bounds, method="Nelder-Mead")
        t = time.time() - t0
        self.logger.info("optimize ksec took %.3g seconds with %d iterations", t, ret.nit)
        if debug:
            ksec_objective(ret.x, debug=True, title="after minimize ksec_objective")
            if self.abort:
                return False
        self.T = ret.x[0]
        self.x0 = ret.x[1]
        self.init_scales = ret.x[2:]    
        self.logger.info("T=%.3g, x0=%.3g, init_scales=%s", self.T, self.x0, str([round(v,3) for v in self.init_scales]))
        return True

    def compute_curve(self, x, x0, scale, rg):
        return lognormal_pore_func(x, scale, self.N, self.T, self.me, self.mp, self.mu, self.sigma, rg, x0)

    def update_column_params(self, params):
        self.N = params[0]
        self.T = params[1]
        self.x0 = params[2]
        self.me = params[3]
        self.mp = params[4]
        self.mu = params[5]
        self.sigma = params[6]

    def plot_state(self, title, lnp_params, save_fig_as=None):
        from molass_legacy.Models.Stochastic.LnporeUtils import plot_lognormal_fitting_state
        ret = plot_lognormal_fitting_state(self.x, self.y, lnp_params, self.rgs, title=title, save_fig_as=save_fig_as)
        if not ret:
            self.abort = True

    def scale_objective_pdf(self, p, debug=False, title=None):
        x = self.x
        y = self.y
        x0 = p[-1]
        cy_list = []
        for rg, scale in zip(self.rgs, p[:-1]):
            cy = lognormal_pore_func(x, scale, self.N, self.T, self.me, self.mp, self.mu, self.sigma, rg, x0)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)

        if debug:
            lnp_params = np.concatenate([self.get_column_params(x0=x0), p[:-1]])
            self.plot_state(title, lnp_params)

        if self.evaluate_area_props:
            props = compute_area_props(cy_list)
            return np.sum((ty - y)**2) + np.sum((props - self.area_props)**2)
        else:
            return np.sum((ty - y)**2)

    def advanced_objective_pdf(self, p, debug=False, title=None, save_fig_as=None):
        x = self.x
        y = self.y
        x0 = p[0]
        mu_ = p[1]
        sigma_ = p[2]
        scales = p[3:]
        cy_list = []
        for rg, scale in zip(self.rgs, scales):
            cy = lognormal_pore_func(x, scale, self.N, self.T, self.me, self.mp, mu_, sigma_, rg, x0)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        if debug:
            print("x0=", x0)
            print("mu, sigma=", mu_, sigma_)
            lnp_params = np.concatenate([self.get_column_params(x0=x0, mu=mu_, sigma=sigma_), scales])
            self.plot_state(title, lnp_params, save_fig_as=save_fig_as)
        return np.sum((y - ty)**2)

    def estimate_psd_with_pdf(self, x, y, rg_recs, area_props, debug=False):
        rgs = rg_recs[:,1]       
        self.logger.info("estimate_psd_with_pdf with x0=%.3g, poresize=%.3g", self.x0, self.poresize)
        self.abort = False
        self.x = x
        self.y = y
        self.rgs = rgs
        self.area_props = area_props

        method = 'Nelder-Mead'
        # method = None

        num_components = len(rgs)
        if self.init_scales is None:
            init_scale = self.area/num_components
            init_params = np.concatenate([np.ones(num_components)*init_scale, [self.x0]])
            if debug:
                self.scale_objective_pdf(init_params, debug=True, title="before scale")
                if self.abort:
                    return

            t0 = time.time()
            min_scale = init_scale*0.2
            scale_bounds = [(min_scale, None)]*(len(init_params)-1) + [(None, None)]
            res = minimize(self.scale_objective_pdf, init_params, method=method, bounds=scale_bounds)
            t = time.time() - t0
            self.logger.info("scale fitting with PDF took : %.3g seconds with %d iterations", t, res.nit)
            if debug:
                print("ret.x=", res.x)
                self.scale_objective_pdf(res.x, debug=True, title="after scale")
                if self.abort:
                    return
                
            init_scales = res.x[:-1]
            init_x0 = res.x[-1]
        else:
            init_scales = self.init_scales
            init_x0 = self.x0

        init_params = np.concatenate([[init_x0, self.mu, self.sigma], init_scales])
        bounds = [(init_x0-50, init_x0+50), (1, 10), (SIGMA_LOWER_BOUND, 1)] + [(0, 10)]*num_components
        if debug:
            self.advanced_objective_pdf(init_params, debug=True, title="before minimize")
            if self.abort:
                return
        t0 = time.time()
        res = minimize(self.advanced_objective_pdf, init_params, method=method, bounds=bounds)
        t = time.time() - t0
        self.logger.info("advanced optimization with PDF took : %.3g seconds with %d iterations", t, res.nit)
        if debug:
            self.advanced_objective_pdf(res.x, debug=True, title="after minimize")
            if self.abort:
                return
        return self.get_whole_params(res)

    def get_whole_params(self, res):
        x0, mu, sigma = res.x[0:3]
        return np.array([self.N, self.T, x0, self.me, self.mp, mu, sigma] + list(res.x[3:]))

    def get_objective_params(self, whole_params):
        return np.concatenate([whole_params[[2,5,6]], whole_params[7:]])

    def compute_residual_ratio(self, x, y, rgs, params, debug=False):
        return compute_residual_ratio_impl(x, y, rgs, params, debug=debug)

    def estimate_psd_with_cf(self, x, y, rg_recs):
        from SecTheory.SecPDF import BidirectFft
        self.logger.info("estimate_psd_with_cf")
        abort = False

        rgs = rg_recs[:,1] 

        def plot_state_from_cf(title, cz_list, tz, mu, sigma):
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
                fig.suptitle(title)
                ax1.plot(x, y)
                for cz in cz_list:
                    cy = bifft.compute_y(cz)
                    ax1.plot(x, cy, ":")
                ty = bifft.compute_y(tz)
                ax1.plot(x, ty, ":", color="red")
                mode = np.exp(mu - sigma**2)
                max_size = int(mode + 500)
                z = np.arange(max_size)
                ax2.plot(z, distr_func(z, mu, sigma), label="pdf")
                fig.tight_layout()
                plt.show()

        bifft = BidirectFft(x)
        z = bifft.compute_z(y)
        w = bifft.get_w()
        def scale_objective(p, debug=False):
            x0 = p[-1]
            cz_list = []
            for rg, scale in zip(rgs, p[:-1]):
                cz = scale*lognormal_pore_cf(w, self.N, self.T, self.me, self.mp, self.mu, self.sigma, rg, x0)
                cz_list.append(cz)
            tz = np.sum(cz_list, axis=0)

            if debug:
                plot_state_from_cf("scale_objective debug", cz_list, tz, self.mu, self.sigma)

            dz = tz - z
            return np.sum(np.real(dz)**2 + np.imag(dz)**2)

        init_params = np.concatenate([np.ones(len(rgs)),[self.x0]])
        scale_objective(init_params, debug=True)
        t0 = time.time()
        res = minimize(scale_objective, init_params, method='Nelder-Mead')
        t = time.time() - t0
        self.logger.info("scale fitting with CF took : %s seconds with %d iterations", t, res.nit)
        scale_objective(res.x, debug=True)
        init_scales = res.x[:-1]
        init_x0 = res.x[-1]

        def advanced_objective(p, debug=False, title=None):
            N = p[0]
            T = p[1]
            x0 = p[2]
            mu_ = p[3]
            sigma_ = p[4]
            scales = p[5:]
            cz_list = []
            for rg, scale in zip(rgs, scales):
                cz = scale*lognormal_pore_cf(w, self.N, self.T, self.me, self.mp, mu_, sigma_, rg, x0)
                cz_list.append(cz)
            tz = np.sum(cz_list, axis=0)

            if debug:
                print("N, T, x0=", N, T, x0)
                print("mu, sigma=", mu_, sigma_)
                plot_state_from_cf(title, cz_list, tz, mu_, sigma_)

            dz = tz - z
            return np.sum(np.real(dz)**2 + np.imag(dz)**2)

        init_params = np.concatenate([[self.N, self.T, init_x0, self.mu, self.sigma], init_scales])
        bounds = [(0, 5000), (0, 10), (-100, 100)] + [(0, 100)]*len(rgs) + [(1, 10), (0.1, 1)]
        advanced_objective(init_params, debug=True, title="before minimize")
        res = minimize(advanced_objective, init_params, method='Nelder-Mead', bounds=bounds)
        self.logger.info("advanced optimization with CF took : %s seconds with %d iterations", t, res.nit)
        advanced_objective(res.x, debug=True, title="after minimize")
        return res.x

    def cover_missing_data(self, rg_list, last_rgs=None):
        if last_rgs is None:
            new_rg_list = rg_list
        else:
            new_rg_list = []
            for rg, last_rg in zip(rg_list, last_rgs):
                if rg is None:
                    new_rg_list.append(last_rg)
                else:
                    new_rg_list.append(rg)
        return np.array(new_rg_list)

def estimate_lognormal_psd_impl(advanced_frame, **kwargs):
    with_cf = kwargs.pop('with_cf', False)
    devel = kwargs.pop('devel',True)
    print("estimate_lognormal_psd_impl: with_cf=", with_cf)

    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    print("estimate_lognormal_psd_impl", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    D, E, qv, ecurve = editor.sd.get_xr_data_separate_ly()
    peak_region = ecurve.get_peak_region(sigma_scale=5)
    print("peak_region=", peak_region)

    def plot_cy_list(cy_list, title, test_y=None):
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":", label="component-%d" % (len(cy_list)))
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label="model total")
            for x_ in peak_region:
                ax.axvline(x=x_+x[0], color="green")
            if test_y is not None:
                ax.plot(x, test_y, ":", color="cyan", label="test curve")
            fig.tight_layout()
            plt.show()

    slice_ = slice(*[int(x_) for x_ in peak_region])
    num_components = len(params_array)
    print("num_components=", num_components)
    M_ = get_denoised_data(D[:,slice_], rank=num_components)
    E_ = E[:,slice_]
    cy_list = compute_cy_list(model, fx, params_array)
    plot_cy_list(cy_list, "initial state")

    min_msd = None
    column = LognormalPoreColumn()
    last_rgs = None
    for k in range(5):
        C = np.array(cy_list)
        C_ = C[:,slice_]
        Cinv = np.linalg.pinv(C_)
        P_ = M_ @ Cinv
        Minv = np.linalg.pinv(M_)
        W = Minv @ P_
        Pe = np.sqrt(E_**2 @ W**2)

        rg_list = []
        for j, p_ in enumerate(P_.T):
            data = np.array([qv, p_, Pe[:,j]]).T
            sg = SimpleGuinier(data)
            rg_list.append(sg.Rg)

        x0 = 300
        scale = 3
        print("rg_list=", rg_list)
        column.prepare_for_rgs(fx, y, rg_list)
        rgs = column.cover_missing_data(rg_list, last_rgs=last_rgs)
        tRs = None
        area_props = compute_area_props(cy_list)

        if with_cf:
            pore_params = column.estimate_psd_with_cf(fx, y, rgs)
        else:
            pore_params = column.estimate_psd_with_pdf(fx, y, rgs, area_props)
        break
        temp_cy_list = column.compute_cy_list(fx, rgs, pore_params)
        ty = np.sum(temp_cy_list, axis=0)
        msd = np.sum((y - ty)**2)
        print("msd=", msd)
        break
        if min_msd is None or msd < min_msd:
            cy_list = temp_cy_list
            min_msd = msd
            column.update_column_params(pore_params)
            plot_cy_list(cy_list, "%d-th estimate" % k)
            last_rgs = rgs
            init_scales = pore_params[9:]
        else:
            break


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

canvas = np.zeros((200, 200))

fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))

n = 120
x = np.linspace(0, np.pi*12, n)

ims = []
for i in range(60):
    t = np.pi*i/3
    beam = 1 + np.sin(x - t)
    canvas[100,0:n] = beam
    im = ax1.imshow(canvas, animated=True)
    ims.append([im])

ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True)
fig.tight_layout()
plt.show()



"""
    Trimming.MomentTrimming.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from .OptViewRange import OptViewRange
from .TrimmingInfo import TrimmingInfo

def moment_trimming_info_debug_plot(plt, uv_x, uv_y, xr_x, xr_y, ranges_xy, ranges):
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    fig.suptitle('Moment Trimming')
    ax1.plot(uv_x, uv_y)
    # ax1.plot(uv_x, baselines[0])
    ax1.plot(*ranges_xy[0], 'o', color='yellow')
    ax1.axvspan(*ranges[0], alpha=0.2)
    ax2.plot(xr_x, xr_y)
    # ax2.plot(xr_x, baselines[1])
    ax2.plot(*ranges_xy[1], 'o', color='yellow')
    ax2.axvspan(*ranges[1], alpha=0.2)
    fig.tight_layout()

def get_info_from_optview_range(uv_x, uv_y, xr_x, xr_y, num_peaks, debug=False):
    ret_info = []
    if debug:
        ranges = []
        ranges_xy = []

    for x, y in [(uv_x, uv_y), (xr_x, xr_y)]:
        range_ = OptViewRange(x, y, num_peaks, upper=0.05)
        lside, rside = range_.get_range()
        pair = [int(max(0, lside)), int(min(rside, len(x)))]
        ret_info.append(TrimmingInfo(1, *pair, len(x)))

        if debug:
            ranges.append(pair)
            ranges_xy.append((range_.x_, range_.y_))
    
    if debug:
        return ret_info, ranges, ranges_xy
    else:
        return ret_info

def get_moment_trimming_info(sd, debug=False, return_debug_info=False):
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xr_curve()
    uv_x = uv_curve.x
    uv_y = uv_curve.y
    xr_x = xr_curve.x
    xr_y = xr_curve.y

    num_peaks = len(xr_curve.peak_info)
    if debug or return_debug_info:
        ret_info, ranges, ranges_xy = get_info_from_optview_range(uv_x, uv_y, xr_x, xr_y, num_peaks, debug=True)
    else:
        ret_info = get_info_from_optview_range(uv_x, uv_y, xr_x, xr_y, num_peaks)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            moment_trimming_info_debug_plot(plt, uv_x, uv_y, xr_x, xr_y, ranges_xy, ranges)
            plt.show()

    if return_debug_info:
        return ret_info, uv_x, uv_y, xr_x, xr_y, ranges, ranges_xy
    else:
        return ret_info

def set_moment_trimming_info(sd):
    from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
    info = get_moment_trimming_info(sd, debug=False)
    uv_restrict_list = get_setting('uv_restrict_list')
    uv_restrict_list[0] = info[0]
    set_setting('uv_restrict_list', uv_restrict_list)
    xr_restrict_list = get_setting('xr_restrict_list')
    xr_restrict_list[0] = info[1]
    set_setting('xr_restrict_list', xr_restrict_list)


"""
    Optimizer.BoundaryView.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

class BoundaryView(Dialog):
    def __init__(self, parent, parent_dialog, state_canvas):
        self.parent = parent
        self.parent_dialog = parent_dialog
        self.elution_model = parent_dialog.elution_model
        self.optimizer = parent_dialog.fullopt
        self.state_canvas = state_canvas
        self.x_array = state_canvas.demo_info[1]
        self.best_index = state_canvas.get_best_index()
        self.fullopt = state_canvas.fullopt
        Dialog.__init__(self, parent, "Boundary View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X, padx=20)
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT)

        self.fig = fig = plt.figure(figsize=(12, 6))
        self.draw_boundary_view(fig)

        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_boundary_view(self, fig):
        job_info = self.parent_dialog.get_job_info()
        job_name = job_info[0]
        in_folder = get_in_folder()

        ax1 = fig.add_subplot(121, projection="3d")
        ax2 = fig.add_subplot(122, projection="polar")
        # fig.tight_layout()
        fig.subplots_adjust(top=0.8)

        fig.suptitle("Parameter Variation Sequence Analysis of Job %s on %s" % (job_name, in_folder), fontsize=20)

        ax1.set_title("Boundary Cylinder 3D View", fontsize=16)
        ax2.set_title("Boundary Cylinder Slice View", fontsize=16)

        """
        Add cylinder to plot
        https://stackoverflow.com/questions/26989131/add-cylinder-to-plot
        """
        # Create the mesh in polar coordinates and compute corresponding Z.
        def data_for_cylinder_along_z(center_x, center_y, radius, sequence_length, nx_array=None):
            num_params = 50 if nx_array is None else nx_array.shape[1]
            z = np.arange(sequence_length)
            theta = np.linspace(0, 2*np.pi, num_params)
            theta_grid, z_grid=np.meshgrid(theta, z)
            if nx_array is None:
                x_grid = radius*np.cos(theta_grid) + center_x
                y_grid = radius*np.sin(theta_grid) + center_y
            else:
                radius_ = nx_array[z,:]
                x_grid = radius_*np.cos(theta_grid) + center_x
                y_grid = radius_*np.sin(theta_grid) + center_y
            return x_grid, y_grid, z_grid

        n_params_list = []
        bounds_mask = self.fullopt.bounds_mask
        lower_bounds = self.fullopt.lower_bounds
        upper_bounds = self.fullopt.upper_bounds
        for params in self.x_array:
            n_params = (params[bounds_mask] - lower_bounds)/(upper_bounds - lower_bounds)
            n_params_list.append(n_params)
        nx_array = np.array(n_params_list)
        sequence_length = nx_array.shape[0]
        print("sequence_length=", sequence_length)
        Xc, Yc, Zc = data_for_cylinder_along_z(0, 0, 1, sequence_length)
        ax1.plot_surface(Zc, Xc, Yc, alpha=0.1)
        Xp, Yp, Zp = data_for_cylinder_along_z(0, 0, 0.5, sequence_length, nx_array)
        ax1.plot_surface(Zp, Xp, Yp, color="green", alpha=0.5)
        ax1.set_box_aspect(aspect = (2,1,1))

        n = 50
        x = np.ones(n) * self.best_index
        theta = np.linspace(0, 2*np.pi, n)
        y = np.sin(theta)
        z = np.cos(theta)
        ax1.plot(x, y, z, color="red", alpha=0.5, lw=3)

        if self.elution_model == 1:
            # stochastic models seem to be inclined to have outliers
            # or, should always do this?
            ax1.set_ylim(-1, 1)
            ax1.set_zlim(-1, 1)

        n_params = nx_array[self.best_index,:]
        n = len(n_params)
        ax2.set_rlim(0, 1)
        ax2.set_xticks(np.pi/180. * np.linspace(0,  360, n, endpoint=False))
        num_all_params = self.x_array.shape[1]
        # all_param_ids = np.arange(num_all_params)
        all_names = self.optimizer.get_parameter_names()
        # print("all_names=", all_names)
        ax2.set_xticklabels(all_names[bounds_mask])
        ax2.set_rlabel_position(180/n)
        theta = np.linspace(0, 2*np.pi, n+1)
        n_params_ = np.concatenate([n_params, n_params[0:1]])
        ax2.plot(theta, n_params_, color="green")



def pairCorrelationFunction_2D(x, y, S, rMax, dr):
    """Compute the two-dimensional pair correlation function, also known
    as the radial distribution function, for a set of circular particles
    contained in a square region of a plane.  This simple function finds
    reference particles such that a circle of radius rMax drawn around the
    particle will fit entirely within the square, eliminating the need to
    compensate for edge effects.  If no such particles exist, an error is
    returned. Try a smaller rMax...or write some code to handle edge effects! ;)

    Arguments:
        x               an array of x positions of centers of particles
        y               an array of y positions of centers of particles
        S               length of each side of the square region of the plane
        rMax            outer diameter of largest annulus
        dr              increment for increasing radius of annulus

    Returns a tuple: (g, radii, interior_indices)
        g(r)            a numpy array containing the correlation function g(r)
        radii           a numpy array containing the radii of the
                        annuli used to compute g(r)
        reference_indices   indices of reference particles
    """
    from numpy import zeros, sqrt, where, pi, mean, arange, histogram
    # Number of particles in ring/area of ring/number of reference particles/number density
    # area of ring = pi*(r_outer**2 - r_inner**2)

    # Find particles which are close enough to the box center that a circle of radius
    # rMax will not cross any edge of the box
    bools1 = x > rMax
    bools2 = x < (S - rMax)
    bools3 = y > rMax
    bools4 = y < (S - rMax)
    interior_indices, = where(bools1 * bools2 * bools3 * bools4)
    num_interior_particles = len(interior_indices)

    if num_interior_particles < 1:
        raise  RuntimeError ("No particles found for which a circle of radius rMax\
                will lie entirely within a square of side length S.  Decrease rMax\
                or increase the size of the square.")

    edges = arange(0., rMax + 1.1 * dr, dr)
    num_increments = len(edges) - 1
    g = zeros([num_interior_particles, num_increments])
    radii = zeros(num_increments)
    numberDensity = len(x) / S**2

    # Compute pairwise correlation for each interior particle
    for p in range(num_interior_particles):
        index = interior_indices[p]
        d = sqrt((x[index] - x)**2 + (y[index] - y)**2)
        d[index] = 2 * rMax

        (result, bins) = histogram(d, bins=edges, density=False)
        g[p, :] = result/numberDensity

    # Average g(r) for all interior particles and compute radii
    g_average = zeros(num_increments)
    for i in range(num_increments):
        radii[i] = (edges[i] + edges[i+1]) / 2.
        rOuter = edges[i + 1]
        rInner = edges[i]
        g_average[i] = mean(g[:, i]) / (pi * (rOuter**2 - rInner**2))

    return (g_average, radii, interior_indices)
####

def pairCorrelationFunction_3D(x, y, z, S, rMax, dr):
    """Compute the three-dimensional pair correlation function for a set of
    spherical particles contained in a cube with side length S.  This simple
    function finds reference particles such that a sphere of radius rMax drawn
    around the particle will fit entirely within the cube, eliminating the need
    to compensate for edge effects.  If no such particles exist, an error is
    returned.  Try a smaller rMax...or write some code to handle edge effects! ;)

    Arguments:
        x               an array of x positions of centers of particles
        y               an array of y positions of centers of particles
        z               an array of z positions of centers of particles
        S               length of each side of the cube in space
        rMax            outer diameter of largest spherical shell
        dr              increment for increasing radius of spherical shell

    Returns a tuple: (g, radii, interior_indices)
        g(r)            a numpy array containing the correlation function g(r)
        radii           a numpy array containing the radii of the
                        spherical shells used to compute g(r)
        reference_indices   indices of reference particles
    """
    from numpy import zeros, sqrt, where, pi, mean, arange, histogram

    # Find particles which are close enough to the cube center that a sphere of radius
    # rMax will not cross any face of the cube
    bools1 = x > rMax
    bools2 = x < (S - rMax)
    bools3 = y > rMax
    bools4 = y < (S - rMax)
    bools5 = z > rMax
    bools6 = z < (S - rMax)

    interior_indices, = where(bools1 * bools2 * bools3 * bools4 * bools5 * bools6)
    num_interior_particles = len(interior_indices)

    if num_interior_particles < 1:
        raise  RuntimeError ("No particles found for which a sphere of radius rMax\
                will lie entirely within a cube of side length S.  Decrease rMax\
                or increase the size of the cube.")

    edges = arange(0., rMax + 1.1 * dr, dr)
    num_increments = len(edges) - 1
    g = zeros([num_interior_particles, num_increments])
    radii = zeros(num_increments)
    numberDensity = len(x) / S**3

    # Compute pairwise correlation for each interior particle
    for p in range(num_interior_particles):
        index = interior_indices[p]
        d = sqrt((x[index] - x)**2 + (y[index] - y)**2 + (z[index] - z)**2)
        d[index] = 2 * rMax

        (result, bins) = histogram(d, bins=edges, density=False)
        g[p,:] = result / numberDensity

    # Average g(r) for all interior particles and compute radii
    g_average = zeros(num_increments)
    for i in range(num_increments):
        radii[i] = (edges[i] + edges[i+1]) / 2.
        rOuter = edges[i + 1]
        rInner = edges[i]
        g_average[i] = mean(g[:, i]) / (4.0 / 3.0 * pi * (rOuter**3 - rInner**3))

    return (g_average, radii, interior_indices)
    # Number of particles in shell/total number of particles/volume of shell/number density
    # shell volume = 4/3*pi(r_outer**3-r_inner**3)
####



# coding: utf-8
"""
    DenoiseFigures.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from mpl_toolkits.mplot3d import Axes3D
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MatrixData import simple_plot_3d
from SvdDenoise import get_denoised_data
from .Qmm3dFigure import find_where_to_zoom_impl

NROWS = 2
NCOLS = 3

class DenoiseFigures(Dialog):
    def __init__(self, parent, dataset, frame):
        self.dataset = dataset
        self.vector = dataset.vector
        self.data = dataset.data
        self.frame = frame
        Dialog.__init__(self, parent, title="Denoise Figures", visible=False)

    def show(self):
        self._show()

    def body(self, bframe):

        fig = plt.figure(figsize=(24,12))
        self.mpl_canvas = FigureCanvasTkAgg(fig, bframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        gs = GridSpec(NROWS,NCOLS)
        axes_list = []
        for i in range(NROWS):
            ax_row = []
            for j in range(NCOLS):
                ax = fig.add_subplot(gs[i,j], projection='3d')
                ax_row.append(ax)
            axes_list.append(ax_row)
        self.axes = np.array(axes_list)

        self.draw_figures()

        fig.tight_layout()

    def draw_figures(self):
        aslice, eslice, w = find_where_to_zoom_impl(self.dataset, self.frame)

        data = self.data[aslice, eslice]
        x = self.vector[aslice]
        y = np.arange(eslice.start, eslice.stop)
        px = x**2

        frame = self.frame

        gy = frame.gy_list[0]
        mu = frame.mu_list[0]
        C = frame.C_list[0]
        print("w=", w, "C.shape=", C.shape)

        Cinv = np.linalg.pinv(C)
        rank = 3

        for i in range(NROWS):
            for j in range(NCOLS):
                ax = self.axes[i,j]
                rank += 3
                M = get_denoised_data(self.data, rank=rank)
                P = np.dot(M, Cinv)
                ax.set_title("Rank=%d" % rank, fontsize=20)
                simple_plot_3d(ax, np.log(M[aslice, eslice]), x=px, y=y, color='green', alpha=0.1)
                for k in w:
                    m = int(round(mu[k]))
                    py = np.ones(len(px))*m
                    z = P[aslice,k]
                    pz = np.log(z*gy[m])
                    ax.plot(px, py, pz, label='c-%d' % k, color='C%d' % (k+2))
                ax.legend()



# coding: utf-8
"""
    ManualDecomposer.py.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy                as np
import logging
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
from molass_legacy.Models.ElutionCurveModels     import EMG

param_names = [ 'h', 'mu', 'sigma', 'tau' ]


class ParamControlPanel:
    def __init__( self, parent, order, pframe, name, init_params ):
        self.parent = parent
        self.order  = order

        frame = Tk.Frame( pframe )
        frame.pack()

        self.init_params = init_params
        self.params = copy.deepcopy( init_params )

        self.scale_list = []

        for k, name in enumerate(param_names):
            init_value = init_params[ name ].value
            from_v  = max( 0.01, init_value * 0.1 )
            to_v    = init_value * 2.0
            if k > 0:
                to_v    = max( 10, to_v )
            resolution = ( to_v - from_v ) / 100

            label = Tk.Label( frame, text=name )
            label.grid( row=k, column=0 )
            var = Tk.DoubleVar()
            scale = Tk.Scale( frame, from_=from_v, to=to_v, resolution=resolution, sliderlength=10, length=200, orient=Tk.HORIZONTAL,
                    command=lambda v, name_=name : self.scale_tracer( v, name_ ) )
            self.scale_list.append( scale )
            scale.grid( row=k, column=1 )
            scale.set( init_value )
            # var.trace( 'w', lambda *args, k_=k, scale_=scale : self.scale_tracer( k_, scale_ ) )

    def scale_tracer( self, v, name ):
        print( 'v=', v, 'name=', name )
        self.params[name].value = float( v )
        self.parent.draw()

    def reset( self ):
        self.params = copy.deepcopy( self.init_params )

class ManualDecomposer( Dialog ):
    def __init__( self, parent, title, y, peak_info, model ):
        self.logger  = logging.getLogger( __name__ )
        self.parent = parent
        self.title_ = title
        self.y  = y
        self.x  = np.arange( len(y) )
        self.peak_info = peak_info
        self.model  = model

    def show( self ):
        Dialog.__init__( self, self.parent, self.title_ )

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self )

        cframe = Tk.Frame( body_frame )
        cframe.pack( side=Tk.LEFT )
        ctlframe = Tk.Frame( body_frame )
        ctlframe.pack( side=Tk.LEFT  )

        emg_model = EMG()

        gs = gridspec.GridSpec( 4, 1 )
        figsize = ( 10, 8 ) if is_low_resolution() else ( 12, 10 )
        fig = plt.figure( figsize=figsize )
        self.ax1 = fig.add_subplot( gs[0:3, 0] )
        self.ax2 = fig.add_subplot( gs[3, 0] )

        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        num_panels = 2

        fit_params = self.get_params()
        h   = fit_params['h']
        fit_params['h'].value = h/num_panels
        self.panels = []
        for k in range(num_panels):
            fit_params_ = copy.deepcopy( fit_params )
            panel = ParamControlPanel( self, k, ctlframe, 'Peak%d' % k, fit_params_ )
            self.panels.append( panel )

        self.draw()

        self.fig.tight_layout()
        self.mpl_canvas.draw()

        self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
        self.toolbar.update()
        # self.mpl_canvas.mpl_connect( 'button_press_event', self.manual_decomposer )

        self.reset_btn  = Tk.Button( ctlframe, text="Reset", command=self.reset )
        self.reset_btn.pack()

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def get_params( self ):
        info = self.peak_info[0]
        start   = info[0]
        head    = int(info[1]+0.5)
        stop    = info[2] + 1
        slice_  = slice( start, stop )
        y_  = self.y[slice_]
        x_  = self.x[slice_]
        params  = self.model.guess(y_, x=x_)
        try:
            out = self.model.fit(y_, params, x=x_)
        except:
            # EMG model simetimes raises ValueError("The input contains nan values")
            self.logger.warning( "%s.fit default method failed. resorting to 'least_squares'." % ( model.name ) )
            out = model.fit(y_, params, x=x_, method='least_squares')

        print( out.params )
        return out.params

    def draw( self ):

        self.ax1.cla()
        self.ax2.cla()

        self.ax1.plot( self.y, color='orange' )

        y_ = copy.deepcopy( self.y )

        my = np.zeros( len(y_) )

        for panel in self.panels:

            ey = self.model.eval( panel.params, x=self.x )
            self.ax1.plot( ey )
            y_ -= ey
            my += ey

        self.ax1.plot( my, ':', color='red' )

        self.ax2.plot( y_, color='gray' )

        ymin, ymax = self.ax1.get_ylim()
        for info in self.peak_info:
            f = info[0]
            t = info[2]
            p = mpl_patches.Rectangle(
                    (f, ymin),      # (x,y)
                    t - f,          # width
                    ymax - ymin,    # height
                    facecolor   = 'cyan',
                    alpha       = 0.2,
                )
            self.ax1.add_patch( p )

        self.mpl_canvas.draw()

    def reset( self ):
        for panel in self.panels:
            panel.reset()
        self.draw()



# -*- coding: utf-8 -*-
"""

    ファイル名：   PilatusImageProperty.py

    処理内容：

        TIFF ファイルヘッダ情報の取得と表示

"""
from __future__ import division, print_function, unicode_literals

import sys
import os
import re
from PIL            import Image, TiffImagePlugin
from molass_legacy.KekLib.OurTkinter     import Tk, tk_set_icon_portable
import OurMessageBox    as MessageBox

IMAGEDESCRIPTION = TiffImagePlugin.IMAGEDESCRIPTION     # 270

def get_tiffinfo( image ):
    info = TiffImagePlugin.ImageFileDirectory()

    description = image.tag.get( IMAGEDESCRIPTION )
    # returned as tuple
    # print 'type(prop_text)=', type(prop_text), '; prop_text=', prop_text
    if description:
        # TODO: len( description ) > 1
        info[IMAGEDESCRIPTION] = description[0]

    return info

def get_propertyies( path ):
    im = Image.open( path )

    tiffinfo =  get_tiffinfo( im )
    description = tiffinfo.get( IMAGEDESCRIPTION )
    # returned again as tuple! ?

    if description:
        return description[0]
    else:
        return 'Tiff file has no property text.'

class ImagePropertyWindow( Tk.Toplevel ):

    def __init__( self, parent, title, path ):
        Tk.Toplevel.__init__( self )

        tk_set_icon_portable( self, 'synthesizer' )

        self.title( title )

        text = get_propertyies( path )
        # print 'DEBUG: text="', text, '"'

        t = Tk.Text( self, setgrid=1, height=20 )
        t.pack( side=Tk.TOP, expand=Tk.Y, fill=Tk.BOTH)
        t.insert('0.0', text )

        w = Tk.Button( self, text="OK", width=5, command=self.ok, default=Tk.ACTIVE )
        w.pack(side=Tk.TOP, padx=5, pady=5)

        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))

        self.update()

    def ok( self ):
        self.destroy()



"""
    Models.Stochastic.MonoporeGuess.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.ModelUtils import compute_raw_moments1, compute_area_props
from molass_legacy.Models.Stochastic.ParamLimits import MAX_PORESIZE

EVALUATE_MOMENT3 = False
M1_WEIGHT = 9
M2_WEIGHT = 1
M3_WEIGHT = 0

def guess_monopore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, monopore_params=None, debug=False):
    N, T, x0, me, mp, poresize = monopore_params[0:6]

    def column_objective(p, debug=False):
        N_, T_, x0_, poresize_ = p
        rhov = peak_rgs/poresize_
        rhov[rhov > 1] = 1
        dev_list = []
        for k, (M, rho) in enumerate(zip(egh_moments_list, rhov)):
            M1_ = x0_ + N_ * T_ * (1 - rho)**(me + mp)
            M2_ = 2 * N_ * T_**2 * (1 - rho)**(me + 2*mp)
            if EVALUATE_MOMENT3:
                M3_ = 6 * N * T**3 * (1 - rho)**(me + 3*mp)
                dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2 + M3_WEIGHT*(M3_ - M[2])**2)
            else:
                dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if debug:
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
                if EVALUATE_MOMENT3:
                    print([k], "M3, M3_ = %.3g, %.3g" % (M[2], M3_))
        return np.sum(np.asarray(dev_list)*props)

    init_params = [N, T, x0, poresize]
    if debug:
        print("init_params=", init_params)
        print("before minimize: fv=", column_objective(init_params, debug=True))

    bounds = [(300, 5000), (0.01, 10), (x0-50, x0+50), (0, MAX_PORESIZE)]
    res = minimize(column_objective, init_params, bounds=bounds, method='Nelder-Mead')
    if debug:
        print("res.x=", res.x)
        print("after minimize: fv=", column_objective(res.x, debug=True))

    N, T, x0, poresize = res.x

    use_nt_ratio = False 
    use_multi_factor_fv = False
    if use_multi_factor_fv:
        raw_moments = np.array([M[0] for M in egh_moments_list])

    def x0_scales_objective(p, title=None):
        if use_nt_ratio:
            K, ratio, x0_ = p[0:3]  # K = N*T, ratio = N/T
            N_ = np.sqrt(K*ratio)   # T = N/ratio, K = N*T = N**2 / ratio
            T_ = np.sqrt(K/ratio)   # N = T*ratio, K = N*T = T**2 * ratio
        else:
            N_, T_, x0_ = p[0:3]
        scales_ = p[3:]
        rhov = peak_rgs/poresize
        rhov[rhov > 1] = 1
        cy_list = []
        for k, (rho, scale) in enumerate(zip(rhov, scales_)):
            ni_ = N_*(1 - rho)**me
            ti_ = T_*(1 - rho)**mp
            cy = scale*robust_single_pore_pdf(x - x0_, ni_, ti_)
            cy_list.append(cy)    
        ty = np.sum(cy_list, axis=0)
        if use_multi_factor_fv:
            raw_moments_ = compute_raw_moments1(x, cy_list)
            props_ = compute_area_props(cy_list)

        if title is not None:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y, 'o')
                for cy in cy_list:
                    ax.plot(x, cy, '-')
                ax.plot(x, ty, ':', color="red")
                fig.tight_layout()
                plt.show()
        if use_multi_factor_fv:
            dev = np.sum((ty - y)**2)**2 + 0.00005*np.sum((raw_moments_ - raw_moments)**2)**2 + 0.00005*np.sum((props_ - props)**2)**2
        else:
            dev = np.sum((ty - y)**2)
        return dev

    num_components = len(peak_rgs)
    scales = monopore_params[6:]
    if use_nt_ratio:
        K = N*T
        ratio = N/T
        init_params = np.concatenate([[K, ratio, x0], scales])
    else:
        init_params = np.concatenate([[N, T, x0], scales])
    if debug:
        print("init_params=", init_params)
        print("before minimize: fv=", x0_scales_objective(init_params, title="before x0_scales_objective minimize"))

    bounds = [(300, 5000), (0.01, 10), (x0-50, x0+50)] + [(0, 10)]*num_components
    res = minimize(x0_scales_objective, init_params, method='Nelder-Mead', bounds=bounds)
    if debug:
        print("res.x=", res.x)
        print("after minimize: fv=", x0_scales_objective(res.x, title="after x0_scales_objective minimize"))
    if use_nt_ratio:
        K, ratio, x0 = res.x[0:3]
        N = np.sqrt(K*ratio)
        T = np.sqrt(K/ratio)
    else:
        N, T, x0 = res.x[0:3]
    scales = res.x[3:]
    monopore_params = np.concatenate([[N, T, x0, me, mp, poresize], scales])
    return monopore_params



# coding: utf-8
"""
    Rgg.KernelDensityDemo.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from pomegranate.distributions import (KernelDensity, UniformKernelDensity as UniformKernelDensityOrig)

class UniformKernelDensity(UniformKernelDensityOrig):
    def __init__(self,  points=[], bandwidth=1, weights=None, frozen=False):
        super(self.__class__, self).__init__()
        self.points = points
        self.bandwidth = bandwidth
        self.weights = weights
        self.frozen = frozen
        self.scale_fix = 1/(self.bandwidth * 2 * len(self.points))

    def probability(self, X):
        p = UniformKernelDensityOrig.probability(self, X)
        return p * self.scale_fix



# coding: utf-8
"""
    ElutionAutoencoder.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
from keras import regularizers
from keras.layers import Input, Dense
from keras.models import Model
from keras import backend as K
import numpy as np 
import matplotlib.pyplot    as plt

DEBUG   = False

USE_LINEAR_ACTIVATION   = True
USE_DEEP_CODERS         = True

def plot_loss( history ):
    #plot our loss 
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model train vs validation loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'validation'], loc='upper right')
    plt.show()

def l1_reg(weight_matrix):
    return 0.01 * K.sum(K.abs(weight_matrix))

class ElutionAutoencoder:
    def __init__( self, a_curve, array ):

        X   = array
        print( 'X.shape=', X.shape )

        num_factors = len( a_curve.peak_info )

        input_dim = X.shape[1]
        encoding_dim = num_factors
        input_img = Input(shape=(input_dim,))
        if USE_LINEAR_ACTIVATION:
            encoded = Dense(encoding_dim, activation='linear')(input_img)
            # encoded = Dense(encoding_dim, activation='linear', activity_regularizer=l1_reg)(input_img)
            # encoded = Dense(encoding_dim, activation='linear', activity_regularizer=regularizers.l1(10e-5))(input_img)
            decoded = Dense(input_dim, activation='linear')(encoded)
        else:
            encoded = Dense(encoding_dim, activation='relu', activity_regularizer=regularizers.l1(10e-5))(input_img)
            # encoded = Dense(encoding_dim, activation='relu', activity_regularizer=l1_reg)(input_img)
            # encoded = Dense(encoding_dim, activation='relu', activity_regularizer=regularizers.activity_l1(1e-4))(input_img)
            decoded = Dense(input_dim, activation='relu')(encoded)
        autoencoder = Model(input_img, decoded)
        autoencoder.compile(optimizer='adam', loss='mse')
        print(autoencoder.summary())

        history = autoencoder.fit(X, X,
                        epochs=100,
                        batch_size=16,
                        # batch_size=256,
                        shuffle=True,
                        validation_split=0.1,
                        verbose = 0)

        plot_loss( history )

        # use our encoded layer to encode the training input
        encoder = Model(input_img, encoded)
        print(encoder.summary())
        # encoded_input = Input(shape=(encoding_dim,))
        # decoder_layer = autoencoder.layers[-1]
        # decoder = Model(encoded_input, decoder_layer(encoded_input))
        encoded_data = encoder.predict(X)

        fig = plt.figure( figsize=(16,7) )
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)

        for k in range(num_factors):
            ax1.plot(encoded_data[:,k])

        ax2.plot( encoded_data[:,0], encoded_data[:,1], 'o' )

        fig.tight_layout()
        plt.show()



"""
    QuickAnalysis.PreDecomposer.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Batch.FullBatch import FullBatch
from molass_legacy.UV.UvPreRecog import UvPreRecog

USE_LRF = False
if USE_LRF:
    from .ModeledPeaks import plot_curve, get_proportions, PeakRegionLrf
else:
    from .ModeledPeaks import plot_curve, get_proportions, get_peak_region_ends

class PreDecomposer(FullBatch):
    def __init__(self, sd, pre_recog, sd_copy, debug=False):
        print("PreDecomposer.__init__")
        self.logger = logging.getLogger(__name__)
        self.sd = sd
        self.pre_recog = pre_recog
        self.sd_copy = sd_copy
        self.xr_curve = sd.get_xray_curve()
        self.ecurve_info = None
        self.unified_baseline_type = get_setting("unified_baseline_type")
        self.upr = UvPreRecog(sd, pre_recog, debug=debug)
        self.base_curve_info = self.upr.get_base_curve_info()

    def update_sd_copy(self, sd_copy):
        self.sd_copy = sd_copy

    def decompose(self, debug=False):
        uv_x, uv_y, xr_x, xr_y, baselines = self.get_curve_xy(return_baselines=True)
        uv_y_ = uv_y - baselines[0]
        xr_y_ = xr_y - baselines[1]
        num_peaks = len(self.xr_curve.peak_info) + 2
        uv_peaks, xr_peaks = self.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_, num_peaks=num_peaks)

        if USE_LRF:
            D, E, qv, _ = self.sd.get_xr_data_separate_ly()
            lrf = PeakRegionLrf(D, E, xr_x, xr_peaks)
            f, t = lrf.ends
            rgs = lrf.compute_rgs(qv)
            print("rgs=", rgs)
        else:
            f, t = get_peak_region_ends(xr_x, xr_peaks)

        if debug:
            from matplotlib.patches import Rectangle
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("decompose: debug", fontsize=20)

                ax1.set_title("UV Elution", fontsize=16)
                ax2.set_title("XR Elution", fontsize=16)

                plot_curve(ax1, uv_x, uv_y, uv_peaks, color='blue', baseline=baselines[0])
                plot_curve(ax2, xr_x, xr_y, xr_peaks, color='orange', baseline=baselines[1])

                ymin, ymax = ax2.get_ylim()
                p = Rectangle(
                        (f, ymin),      # (x,y)
                        t - f,          # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax2.add_patch(p)

                fig.tight_layout()
                plt.show()

        self.uv_x = uv_x
        self.xr_x = xr_x
        self.uv_peaks = uv_peaks
        self.xr_peaks = xr_peaks

    def get_xr_proportions(self):
        return get_proportions(self.xr_x, self.xr_peaks)

def spike(caller):
    from time import time
    from importlib import reload
    import QuickAnalysis.Homopeak
    reload(QuickAnalysis.Homopeak)
    from .Homopeak import Homopeakness

    t0 = time()
    sd = caller.sd_orig
    sd_copy = caller.serial_data
    pre_recog = caller.pre_recog_orig
    predec = PreDecomposer(sd, pre_recog, sd_copy, debug=False)
    predec.decompose(debug=True)
    hpn = Homopeakness(predec)
    hpn.get_homopeakness_scores()
    print("it took ", time() - t0)



"""
    ABC.CustomSamplers.py

    see pyabc.platform_factory.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from typing import Union
from jabbar import jabbar
from pyabc.sampler import SingleCoreSampler

class DefaultSampler(SingleCoreSampler):
    def __init__(self, check_max_eval: bool = False):
        from importlib import reload
        import Solvers.ABC.CustomPopulation
        reload(Solvers.ABC.CustomPopulation)
        from Solvers.ABC.CustomPopulation import CustomSampleFactory

        self.nr_evaluations_: int = 0
        self.sample_factory: CustomSampleFactory = CustomSampleFactory(
            record_rejected=False
        )
        self.show_progress: bool = False
        self.analysis_id: Union[str, None] = None

        self.check_max_eval = check_max_eval

    def sample_until_n_accepted(
        self,
        n,
        simulate_one,
        t,
        *,
        max_eval=np.inf,
        all_accepted=False,
        ana_vars=None,
    ):
        nr_simulations = 0
        sample = self._create_empty_sample()

        for _ in jabbar(range(n), enable=self.show_progress, keep=False):
            while True:
                if self.check_max_eval and nr_simulations >= max_eval:
                    break
                new_sim = simulate_one()
                sample.append(new_sim)
                nr_simulations += 1
                if new_sim.accepted:
                    break
        self.nr_evaluations_ = nr_simulations

        if sample.n_accepted < n:
            sample.ok = False

        return sample


"""
    Models.Stochastic.ExclusionCurves.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from molass_legacy.Models.Stochastic.ParamLimits import USE_K

RESOL_C = 100
RESOL_F = 1000
def get_resolusin(fine=False):
    return RESOL_F if fine else RESOL_C

def simple_curve(rv, N, T, t0, m, poresize):
    rhov = rv/poresize
    tv = t0 + N*T*(1 - rhov)**m
    return tv

def compute_simply_conformant_rgs(N, KT, t0, poresize, m, trs, debug=False, rgs=None, plot_info=None):
    from importlib import reload
    import Models.Stochastic.MonoporeUtils as module
    reload(module)
    from molass_legacy.Models.Stochastic.MonoporeUtils import plot_monopore_conformance

    if USE_K:
        T = KT/N
    else:
        T = KT
    # trs = t0 + N*T*(1 - rhov)**m
    # (trs - t0)/(N*T) = (1 - rhov)**m
    small_values = np.ones(len(trs))
    tr_diff = np.max([trs - t0, small_values], axis=0)
    rhov = 1 - np.power(tr_diff/(N*T), 1/m)
    conformant_rgs = rhov * poresize

    if debug and plot_info is not None:
        plot_monopore_conformance(KT, t0, poresize, plot_info, rg_info=(rgs, trs), confrg_info=(conformant_rgs, trs))

    return conformant_rgs

def plot_simple_curves(ax, N, T, t0, m, poresize_list):
    ax.set_title("Simple Exclusion Curves with N=%d, T=%.1f, t0=%.1f, m=%.1f" % (N, T, t0, m))
    for poresize in poresize_list:
        rv = np.linspace(poresize, 10, 100)
        tv = simple_curve(rv, N, T, t0, m, poresize)
        ax.plot(tv, rv, label="poresize=%d" % poresize)
    ax.axvline(x=t0, color="red", label="$t_0$=%d" % t0)
    ax.legend()

def monopore_curve(rv, N, T, t0, me, mp, poresize, fine=False):
    from SecTheory.BasicModels import robust_single_pore_pdf
    rhov = rv/poresize
    niv = N*(1 - rhov)**me
    tiv = T*(1 - rhov)**mp
    x = np.linspace(t0, N*T, get_resolusin(fine))
    tv = []
    for ni, ti in zip(niv, tiv):
        y = robust_single_pore_pdf(x - t0, ni, ti)
        m = np.argmax(y)
        tv.append(x[m])
    return np.asarray(tv)

def plot_monopore_curves(ax, N, T, t0, m, poresize, fine=False):
    ax.set_title("Monopore Exclusion Curves with N=%d, T=%.3g, poresize=%g" % (N, T, poresize))
    rv = np.linspace(poresize, 10, get_resolusin(fine))
    tv = simple_curve(rv, N, T, t0, m, poresize)
    ax.plot(tv, rv, label="simple, m=%g" % m)
    for me in [1, 1.5, 2]:
        mp = m - me
        tv = monopore_curve(rv, N, T, t0, me, mp, poresize, fine=fine)
        ax.plot(tv, rv, label="me, mp=(%g, %g)" % (me, mp))
    ax.axvline(x=t0, color="red", label="$t_0$=%d" % t0)
    ax.legend()

def lnpore_curve(rv, N, T, t0, me, mp, mu, sigma, fine=False):
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_func
    x = np.linspace(t0, N*T, get_resolusin(fine))
    tv = []
    for rg in rv:
        y = lognormal_pore_func(x, 1, N, T, me, mp, mu, sigma, rg, t0)
        m = np.argmax(y)
        tv.append(x[m])
    return np.asarray(tv)

def plot_lnpore_curves(ax, N, T, t0, m, poresize, fine=False):
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma
    ax.set_title("Lnpore Exclusion Curves with N=%d, T=%.3g, poresize=%g" % (N, T, poresize))
    rv = np.linspace(poresize, 10, get_resolusin(fine))
    tv = simple_curve(rv, N, T, t0, m, poresize)
    ax.plot(tv, rv, label="simple, m=%g" % m)
    me = 1.5
    mp = 1.5
    for stdev_ratio in [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]:
        stdev = stdev_ratio*poresize
        mu, sigma = compute_mu_sigma(poresize, stdev)
        tv = lnpore_curve(rv, N, T, t0, me, mp, mu, sigma, fine=fine)
        ax.plot(tv, rv, label=r"$\mu$, $\sigma$=(%.3g, %.3g)" % (mu, sigma))
    ax.axvline(x=t0, color="red", label="$t_0$=%d" % t0)
    ax.legend()

def demo(N, T, t0, m, fine=False):
    from time import time
    start_t = time()
    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,8))
    fig.suptitle("Exclusion Curves: Monopore vs. Lognormalpore", fontsize=20)
    plot_simple_curves(axes[0,0], N, T, t0, m, [30, 50, 100, 200, 300])
    plot_monopore_curves(axes[0,1], N, T, t0, m, 100, fine=fine)
    plot_lnpore_curves(axes[1,0], 4000, 0.25, t0, m, 100, fine=fine)
    plot_lnpore_curves(axes[1,1], N, T, t0, m, 100, fine=fine)
    fig.tight_layout()
    print("Elapsed time: %.3f sec" % (time() - start_t))
    plt.show()

if __name__ == "__main__":
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")

    demo(2000, 0.5, 100, 3, fine=False)     # it takes 10 min when fine=True


"""
    MCMC.SolverEmcee.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import emcee
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.StateSequence import save_opt_params
from molass_legacy.Optimizer.OptimizerUtils import OptimizerResult

def walkers_independent(coords):
    if not np.all(np.isfinite(coords)):
        return False
    C = coords - np.mean(coords, axis=0)[None, :]
    C_colmax = np.amax(np.abs(C), axis=0)
    if np.any(C_colmax == 0):
        return False
    C /= C_colmax
    C_colsum = np.sqrt(np.sum(C**2, axis=0))
    C /= C_colsum
    return np.linalg.cond(C.astype(float)) <= 1e8

class SolverEmcee:
    def __init__(self, optimizer):
        self.optimizer = optimizer
        self.cb_fh = optimizer.cb_fh
        self.nfev = 0
        self.logger = logging.getLogger(__name__)

    def log_likelihood(self, theta):
        self.nfev += 1
        return -self.objective(theta)

    def log_prior(self, theta):
        if np.min(theta - self.bounds[:,0]) >= 0 and np.max(theta - self.bounds[:,1]) <= 0:
            return 0.0
        return -np.inf

    def log_probability(self, theta):
        lp = self.log_prior(theta)
        if not np.isfinite(lp):
            return -np.inf
        return lp + self.log_likelihood(theta)

    def minimize(self, objective, init_params, niter=100, seed=1234, bounds=None, callback=None, debug=True):
        self.objective = objective
        self.bounds = bounds
        if callback is None:
            callback = self.callback
        ndim = len(init_params)
        nwalkers = ndim*4
        temp_params = init_params.copy()
        is_nan = np.isnan(temp_params)
        temp_params[is_nan] = np.mean(bounds[is_nan,:], axis=1)

        if debug:
            self.optimizer.debug_plot_params(temp_params)

        callback(temp_params, 0, False)
        p0 = [temp_params + 1e-7*np.random.randn(ndim) for i in range(nwalkers)]
        ret = walkers_independent(p0)
        print("ret=", ret)
        self.sampler = sampler = emcee.EnsembleSampler(nwalkers, ndim, self.log_probability, args=())

        self.logger.info("Running burn-in...")
        p0, _, _ = sampler.run_mcmc(p0, 100)

        self.logger.info("Running production...")
        for sample in sampler.sample(p0, iterations=200):
            # Only check convergence every 100 steps
            if sampler.iteration % 10:
                continue

            samples = sampler.flatchain
            theta_max = samples[np.argmax(sampler.flatlnprobability)]
            theta_max = sampler.flatchain[np.argmax(sampler.flatlnprobability)]
            callback(theta_max, 0, True)

        return OptimizerResult(x=theta_max, nit=niter, nfev=self.nfev)
    
    def callback(self, norm_params, f, accept):
        fv = self.objective(norm_params)
        self.logger.info("callback: fv=%.3g", fv)
        real_params = self.optimizer.to_real_params(norm_params)
        save_opt_params(self.cb_fh, real_params, fv, accept, self.nfev)
        return False


"""
    PyChemLib3D.py

    functions borrowed from
    https://pythoninchemistry.org/sim_and_scat/intro

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from scipy.constants import Boltzmann
mass_of_argon = 39.948 # amu

from .PyChemLib import lj_force

NDIM = 3
BOX_LIMITS = np.array([(-40, 40), (-40, 40), (-40, 40)])

def init_velocity(T, num_particles, mass):
    """
    Initialise the velocities for a series of 
    particles.
    
    Parameters
    ----------
    T: float
        Temperature of the system at 
        initialisation (K)
    number_of_particles: int
        Number of particles in the system
    
    Returns
    -------
    ndarray of floats
        Initial velocities for a series of 
        particles (eVs/Åamu)
    """
    R = np.random.rand(num_particles, NDIM) - 0.5
    return R * np.sqrt(Boltzmann * T / (
        mass * 1.602e-19))

def get_accelerations(positions, mass, epsilon, sigma, closure_radius):
    """
    Calculate the acceleration on each 
    particle as a  result of each other 
    particle. 
    N.B. We use the Python convention of
    numbering from 0.
    
    Parameters
    ----------
    positions: ndarray of floats
        The positions, in a single dimension, 
        for all of the particles (Å)
        
    Returns
    -------
    ndarray of floats
        The acceleration on each particle (eV/Åamu)
    """
    n = positions.shape[0]
    dim = positions.shape[1]
    assert dim == NDIM
    accel_x = np.zeros((n, n, dim))
    for i in range(0, n - 1):
        for j in range(i + 1, n):
            r_x = positions[j] - positions[i]
            rmag = np.sqrt(np.sum(r_x**2))
            force_scalar = lj_force(max(1, rmag - closure_radius), epsilon, sigma)
            force_x = force_scalar * r_x / rmag
            accel_x[i, j] = force_x / mass  # eV Å-1 amu-1
            # appling Newton's third law
            accel_x[j, i] = - force_x / mass
    return np.sum(accel_x, axis=0)

def demo_accel():
    mass_of_argon = 39.948 # amu
    epsilon = 0.0103
    sigma = 3.4
    accel = get_accelerations(np.array([(1,1,1), (5,5,5), (10,10, 10)]), mass_of_argon, epsilon, sigma)
    print('Acceleration on particle 0 = %.3e, %.3e, %.3e eV/Åamu' % tuple(accel[0]))
    print('Acceleration on particle 1 = %.3e, %.3e, %.3e eV/Åamu' % tuple(accel[1]))
    print('Acceleration on particle 2 = %.3e, %.3e, %.3e eV/Åamu' % tuple(accel[2]))

def update_pos(x, v, a, dt, box_limits):
    """
    Update the particle positions.
    
    Parameters
    ----------
    x: ndarray of floats
        The positions of the particles in a 
        single dimension
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New positions of the particles in a single 
        dimension
    """
    zzzz = np.zeros(len(x))
    xmin = np.ones(len(x))*box_limits[0][0]
    xmax = np.ones(len(x))*box_limits[0][1]
    ymin = np.ones(len(x))*box_limits[1][0]
    ymax = np.ones(len(x))*box_limits[1][1]
    x_ = x + v * dt + 0.5 * a * dt * dt
    exmin = np.min([x_[:,0] - xmin, zzzz], axis=0)
    exmax = np.max([x_[:,0] - xmax, zzzz], axis=0)
    eymin = np.min([x_[:,1] - ymin, zzzz], axis=0)
    eymax = np.max([x_[:,1] - ymax, zzzz], axis=0)
    ezmin = np.min([x_[:,2] - ymin, zzzz], axis=0)
    ezmax = np.max([x_[:,2] - ymax, zzzz], axis=0)
    ex = exmin + exmax
    ey = eymin + eymax
    ez = ezmin + ezmax
    e_ = np.array([ex, ey, ez]).T
    v[ex!=0,0] *= -1
    v[ey!=0,1] *= -1
    v[ez!=0,2] *= -1
    return x_ - 2*e_

def update_velo(v, a, a1, dt):
    """
    Update the particle velocities.
    
    Parameters
    ----------
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension (eVs/Åamu)
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension at the previous 
        timestep (eV/Åamu)
    a1: ndarray of floats
        The accelerations of the particles in a
        single dimension at the current 
        timestep (eV/Åamu)
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New velocities of the particles in a
        single dimension (eVs/Åamu)
    """
    return v + 0.5 * (a + a1) * dt

def run_md(num_particles, dt, number_of_steps, initial_temp, x, mass, box_limits, epsilon, sigma, closure_radius):
    """
    Run a MD simulation.
    
    Parameters
    ----------
    dt: float
        The timestep length (s)
    number_of_steps: int
        Number of iterations in the simulation
    initial_temp: float
        Temperature of the system at 
        initialisation (K)
    x: ndarray of floats
        The initial positions of the particles in a 
        single dimension (Å)
        
    Returns
    -------
    ndarray of floats
        The positions for all of the particles 
        throughout the simulation (Å)
    """
    positions = np.zeros((number_of_steps, num_particles, NDIM))
    v = init_velocity(initial_temp, num_particles, mass)
    a = get_accelerations(x, mass, epsilon, sigma, closure_radius)
    for i in range(number_of_steps):
        x = update_pos(x, v, a, dt, box_limits)
        a1 = get_accelerations(x, mass, epsilon, sigma, closure_radius)
        v = update_velo(v, a, a1, dt)
        a = np.array(a1)
        positions[i, :] = x
    return positions

def generate_ensemble_impl(num_particles, box_limits):
    ndim = len(box_limits)
    points = np.random.uniform(0, 1, (num_particles, ndim))
    for i, (vmin, vmax) in enumerate(box_limits):
        points[:,i] = vmin + points[:,i]*(vmax - vmin)
    return points

def generate_ensemble(num_particles, box_limits, min_dist=None):
    points = generate_ensemble_impl(num_particles, box_limits)
    if min_dist is None:
        return points

    min_dist2 = min_dist**2
    while True:
        too_close = False
        for i in range(num_particles):
            for j in range(i+1, num_particles):
                dist2 = np.sum((points[i] - points[j])**2)
                if dist2 < min_dist2:
                    too_close = True
                    break
            if too_close:
                break
        if too_close:
            print("retrying due to too close", np.sqrt(dist2))
            points = generate_ensemble_impl(num_particles, box_limits)
            continue
        else:
            break

    return points

def compute_min_dist(points):
    num_particles = len(points)
    min_dist2 = None
    for i in range(num_particles):
        for j in range(i+1, num_particles):
            dist2 = np.sum((points[i] - points[j])**2)
            if min_dist2 is None or dist2 < min_dist2:
                min_dist2 = dist2
    return np.sqrt(min_dist2)

def demo_lj():
    epsilon = 0.0103
    sigma = 3.4
    r = np.linspace(3.5, 8, 100)

    fig, ax1 = plt.subplots()
    ax1.set_title("Lennard-Jones Potential")
    ax1.plot(r, lj_force(r, epsilon, sigma))
    ax1.set_xlabel("Distance")
    ax1.set_ylabel("Potential")
    fig.tight_layout()
    plt.show()

def demo_run_md(num_particles):
    epsilon = 0.0103
    sigma = 3.4
    closure_radius = 4.5

    box_limits = BOX_LIMITS
    x = generate_ensemble(num_particles, box_limits, min_dist=6)
    num_particles = len(x)
    sim_pos = run_md(num_particles, 0.1, 20000, 300, x, mass_of_argon, box_limits, epsilon, sigma, closure_radius)
    num_frames = len(sim_pos)//100

    dist_array = np.zeros(num_frames)
    for k in range(num_frames):
        dist_array[k] = compute_min_dist(sim_pos[k*100])
    max_i = np.argmax(dist_array)

    pause = False

    def on_click(event):
        nonlocal pause
        print('on_click')
        if event.inaxes is None:
            return
        pause ^= True

    fig, ax2 = plt.subplots(subplot_kw=dict(projection='3d'))
    cid = fig.canvas.mpl_connect('button_press_event', on_click)

    ax2.set_title("Particle Movement Animation N=%d" % num_particles)
    iter_text = ax2.text2D(0.1, 0.9, "Iter: 0", alpha=0.5, transform=ax2.transAxes)
    dist_text = ax2.text2D(0.75, 0.9, "Dist: _", alpha=0.5, transform=ax2.transAxes)
    point_arts, = ax2.plot(sim_pos[0,:,0], sim_pos[0,:,1], sim_pos[0,:,2], "o")
    ax2.set_xlabel(r'X')
    ax2.set_ylabel(r'Y')
    ax2.set_zlabel(r'Z')
    ax2.set_xlim(*box_limits[0])
    ax2.set_ylim(*box_limits[1])
    ax2.set_zlim(*box_limits[2])
    # ax2.legend()
    fig.tight_layout()

    def index_generator():
        i = 0
        while i < num_frames:
            i_ = i
            if pause:
                pass
            else:
                i += 1
            yield i_

    def init():
        iter_text.set_text("Iter: 0")
        dist_text.set_text("Dist: %.1f" % dist_array[0])
        return point_arts, iter_text, dist_text

    def update(i):
        pos = sim_pos[i*100]
        point_arts.set_data(pos[:,0], pos[:,1])
        point_arts.set_3d_properties(pos[:,2])
        iter_text.set_text("Iter: %d" % i)
        dist_text.set_text("Dist: %.1f" % dist_array[i])
        return point_arts,iter_text, dist_text

    ani = FuncAnimation(fig, update, index_generator,
                        init_func=init, blit=True)

    plt.show()

    ani.save("md.gif")

    print("max min_dist=%.1f" % dist_array[max_i])
    pos = sim_pos[max_i*100]
    np.savetxt("points.dat", pos)



# coding: utf-8
"""

    HtmlColorNames.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
# HTML Color Names from http://www.w3schools.com/colors/colors_names.asp
black       = '000000'
red         = 'FF0000'
blue        = '0000FF'
gold        = 'FFD700'
orange      = 'FFA500'
green       = '008000'
seagreen    = '2E8B57'
darkseagreen= '8FBC8F'
limegreen   = '32CD32'
lightgreen  = '90EE90'
steelblue   = '4682B4'
deepskyblue = '00BFFF'
lightskyblue= '87CEFA'
dodgerblue  = '1E90FF'
royalblue   = '4169E1'
lightgray   = 'D3D3D3'
darkgray    = 'A9A9A9'
darkorange  = 'FF8C00'

"""
    to convert to int do like as follows

    int_color = int( red, 16 )

    or

    from ExcelCOM import RGB
    xl_rgb_color = RGB( red )

    note that in Excel VBA
        0x50B000 == RGB(0, 176, 80)
"""



"""
   AdhocFigure.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

MODEL_NAME_DICT = {
    "G0346": "EGH",
    "G1100": "SDM",
    "G2010": "EDM",
}

def show_adhoc_figure_impl(js_canvas):
    print("show_adhoc_figure_impl")
    optimizer = js_canvas.fullopt
    func_code = optimizer.get_name()
    model_name = MODEL_NAME_DICT[func_code]
    params = js_canvas.get_current_params()
    job_name = js_canvas.dialog.get_job_info()[0]
    in_folder = get_in_folder()

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("V2 Result at %s with=%s of %s" % (job_name, model_name, in_folder), fontsize=16)
        optimizer.objective_func(params, plot=True, axis_info=(fig, (None, ax, None, None)))
        fig.tight_layout()
        plt.show()


# coding: utf-8
"""
    SaxsCurve.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt

class SaxsCurve:
    def __init__(self, fig, ax, x, in_y, r_y):
        ax.set_title("Scattering Curve", fontsize=20)

        ax.set_xlabel('Q', fontsize=16)
        ax.set_ylabel('Log(I)', fontsize=16)

        scale = 1 if in_y is None else in_y[0]/r_y[0]
        ax.plot(x, np.log10(r_y*scale), label='simulated')
        if in_y is not None:
            ax.plot(x, np.log10(in_y), ':', label='experiment')

        ax.legend(fontsize=16)



"""
    PersistentPy.py - succcesor of PersistentInfo.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF
"""
import os
import shutil
from molass_legacy.KekLib.BasicUtils import home_dir, exe_name
import molass_legacy.KekLib.CustomMessageBox as MessageBox

def kek_tools_folder():
    return '%s/.KekTools' % (home_dir())

def get_py_file_path(name):
    return '%s/%s/%s' % ( kek_tools_folder(), exe_name(), name )

def make_py_file_path(folder, file):
    if not os.path.exists( folder ):
        os.makedirs( folder )
    app_folder   = '%s/%s' % ( folder, exe_name() )
    if not os.path.exists( app_folder ):
        os.makedirs( app_folder )

    file_path = '%s/%s' % ( app_folder, file )
    if os.path.exists(file_path):
        # check if the file is writable
        assert os.access(file_path, os.W_OK)
    else:
        # check if the folder is writable
        try:
            access_test_file = os.path.join(app_folder, 'access_test')
            fh = open(access_test_file, 'w')
            fh.close()
            os.remove(access_test_file)
        except Exception as exc:
            print('access error in', app_folder, exc)
            raise

    return file_path

def load_py_file(filename, locals_={}):
    from molass_legacy.KekLib.EvalUtils import eval_file
    return eval_file(filename, locals_)

class PersistentPy:
    def __init__(self, filename, defaults={}, py_file=None, alt_folder=None, locals_=None, warn_on_fail=True):
        self.alt_folder = alt_folder
        self._dict = defaults
        if py_file is None:
            try:
                if alt_folder is None:
                    folder = kek_tools_folder()
                else:
                    folder = alt_folder

                py_file = make_py_file_path(folder, filename)
                self.alt_folder = alt_folder
            except:
                try:
                    alt_folder = __file__
                    for k in range(3):
                        alt_folder, _ = os.path.split(alt_folder)
                        print([k], alt_folder)
                    alt_folder = os.path.join(alt_folder, '.KekTools')
                    py_file = make_py_file_path(alt_folder, filename)
                    MessageBox.showwarning( 'Alternative Settings Folder Notification',
                        'Using %s\n'
                        'as your settings folder '
                        'due to an unidentified error.\n'
                        % alt_folder
                        )
                    self.alt_folder = alt_folder
                except:
                    MessageBox.showerror( 'No other alternative',
                        'Failed to use %s\n' % alt_folder
                        )
                    assert False

        self.py_file = py_file

        if os.path.exists(self.py_file) and os.path.getsize(self.py_file):
            try:
                self._dict = load_py_file(self.py_file, locals_=locals_)
            except Exception as exc:
                if True:
                    import logging
                    logger = logging.getLogger(__name__)
                    logger.warning("in pid=%d, load_py_file(%s) failed due to %s", os.getpid(), self.py_file, exc)
                else:
                    from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                    etb = log_exception(None, "in pid=%d, load_py_file(%s): " % (os.getpid(), self.py_file), n=10)

                if warn_on_fail:
                    MessageBox.showwarning( 'Broken Settings Notification',
                        'Your settings infomation previously stored in\n'
                        '"%s" seems to have been broken.\n'
                        '\n'
                        'Sorry, your entire settings information will be lost\n'
                        'and reset to the default.\n'
                        '\n'
                        'This inconvenience can happen in an invocation just after such cases as listed.\n'
                        '  - abnormal termination of the program\n'
                        '  - multiple simutaneous invocation of the program\n'
                        % (self.py_file) )
                            # TODO: set parent

                py_dir = os.path.dirname(self.py_file)
                try:
                    # shutil.rmtree(py_dir)
                    pass
                except:
                    # continue anyway
                    pass
                self._dict = {}
        else:
            self._dict = {}

        for k in defaults.keys():
            if self._dict.get(k) == None:
                self._dict[k] = defaults[k]

    def set_dictionary(self, dictionary):
        self._dict = dictionary

    def get_dictionary(self):
        return self._dict

    def save(self, file=None, pprint=True):
        if file is None:
            file = self.py_file
        with open(file, "w") as fh:
            if pprint:
                import pprint
                fh.write(pprint.pformat(self._dict))
            else:
                fh.write(str(self._dict))